/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.impl;

import org.xmloperator.lambda.net.exception.AuxiliaryPortException;
import org.xmloperator.lambda.net.exception.MarkedDifferentlyException;
import org.xmloperator.lambda.net.model.Abstractor;
import org.xmloperator.lambda.net.model.ConnectionListener;
import org.xmloperator.lambda.net.model.Delimiter;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorList;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.ScopeBinder;

/**
 * Delimiter implementation.
 */
final class DelimiterImpl extends OperatorImpl implements Delimiter {
	private static final boolean SCOPE_BINDING = Test.SCOPE_BINDING;

	final PortImpl input;
	final PortImpl body;
	boolean hasOrientationReversed = false;
	int index = 0;
	ScopeBinder scopeBinder = null;

	DelimiterImpl(
		OperatorFactoryImpl operatorFactory,
		int mark,
		boolean isDirectionInversed) {
		super(operatorFactory, mark);
		this.body =
			new PortImpl(Port.TYPE_DELIMITER_BODY, this, 0, isDirectionInversed);
		this.input =
			new PortImpl(Port.TYPE_DELIMITER_INPUT, this, 1, !isDirectionInversed);
	}

	void setDirectionInversed(boolean isDirectionInversed) {
		this.body.isInput = isDirectionInversed;
		this.input.isInput = !isDirectionInversed;
	}

	//////////////////////////
	// Operator implementation
	//////////////////////////

	public boolean canBeReturnedToFactory() {
		return super.canBeReturnedToFactory()
			&& this.input.connectedPort == null
			&& this.body.connectedPort == null;
	}

	public void returnToFactory() {
		super.returnToFactory();
		if (this.scopeBinder != null) {
			OperatorList delimiterList = this.scopeBinder.getDelimiterList();
			delimiterList.remove(this);
			if (delimiterList.isEmpty())
				this.scopeBinder.returnToFactory();
			this.scopeBinder = null;
		}
		this.hasOrientationReversed = false;
		this.index = 0;
		this.body.number = 0;
		this.input.number = 1;
		this.operatorFactory.returnDelimiter(this);
	}

	public int getPortCount() {
		return 2;
	}

	public int getOperatorType() {
		return Operator.TYPE_DELIMITER;
	}

	public boolean isDelimiter() {
		return true;
	}

	public Port getPort(int number) {
		switch (number) {
			case 0 :
				return this.hasOrientationReversed
					? (Port) this.input
					: (Port) this.body;
			case 1 :
				return this.hasOrientationReversed
					? (Port) this.body
					: (Port) this.input;
			case 2 :
				return null;
			default :
				throw new IllegalArgumentException();
		}
	}

	public Port getPort0() {
		return this.hasOrientationReversed ? (Port) this.input : (Port) this.body;
	}

	public Port getPort1() {
		return this.hasOrientationReversed ? (Port) this.body : (Port) this.input;
	}

	public Port getPort2() {
		return null;
	}

	public boolean isDirectionInversed() {
		return this.body.isInput;
	}

	public Operator cloneOperator(boolean isDirectionInversed) {
		DelimiterImpl delimiter =
			(DelimiterImpl) this.operatorFactory.createDelimiter(
				this.mark,
				isDirectionInversed,
				this.scopeBinder);
		if (this.hasOrientationReversed)
			delimiter.reverseOrientation();
		else
			delimiter.index = this.index;
		return delimiter;
	}

	/**
	 * The Delimiter "D" commutes with the operator "B" :
	 * (A--D>--<B--C) -> (A--<B--D>--C)
	 * or, if identical, annihilates it :
	 * (A--D>--<D--C) -> (A--C)
	 * or disappear into an Identity Abstractor or a closed Abstractor :
	 * (A--D>--<I) -> (A--<I)
	 */
	public int interact(ConnectionListener xListener) {
		Port portD0 = this.getPort0();
		Port portB0 = portD0.getConnectedPort();
		Operator operatorB = portB0.getOperator();
		if (operatorB.isEraser())
			return operatorB.interact(xListener);
		if (!portB0.isPrincipal())
			throw new AuxiliaryPortException(portB0);
		if (operatorB.getMark() != this.mark)
			throw new MarkedDifferentlyException(operatorB);
		Port portD1 = this.getPort1();
		Port portA1 = portD1.getConnectedPort();
		portD0.disconnect();
		portD1.disconnect();
		Port portB1 = operatorB.getPort1();
		Port portC1 = portB1.getConnectedPort();
		boolean isIdentity =
			operatorB.isAbstractor() && ((Abstractor) operatorB).isIdentity();
		if (isIdentity
			|| (SCOPE_BINDING
				&& !this.hasOrientationReversed
				&& operatorB.isAbstractor()
				&& ((Abstractor)operatorB).getScopeBinder() == null)) {
			// Operator B == Identity Abstractor or closed Abstractor.
			portA1.connectTo(portB0);
			xListener.notifyConnection(portA1, portB0);
			this.returnToFactory();
			return isIdentity
				? Operator.INTERACTION_EXTRUDE_IDENTITY
				: Operator.INTERACTION_EXTRUDE_CLOSED_ABSTRACTOR;
		}
		portB1.disconnect();
		if (operatorB.isDelimiter()
			&& ((DelimiterImpl) operatorB).index == this.index) {
			// Annihilates.
			portA1.connectTo(portC1);
			xListener.notifyConnection(portA1, portC1);
			operatorB.returnToFactory();
			this.returnToFactory();
			return Operator.INTERACTION_ANNIHILATE_DELIMITER;
		}
		// Commutes.
		portA1.connectTo(portB0);
		xListener.notifyConnection(portA1, portB0);
		portB1.connectTo(portD1); // Not a principal connection.
		portD0.connectTo(portC1);
		xListener.notifyConnection(portD0, portC1);
		// Updates indices.
		if (operatorB.isDelimiter()) {
			DelimiterImpl delimiterB = (DelimiterImpl) operatorB;
			if (this.index < delimiterB.index)
				delimiterB.incrementIndex();
			else
				this.incrementIndex();
		}
		else if (operatorB.isAbstractor())
			this.incrementIndex();
		else if (operatorB.isDuplicator()) {
			DuplicatorImpl duplicatorB = (DuplicatorImpl) operatorB;
			if (duplicatorB.index >= this.index)
				duplicatorB.incrementIndex();
		}
		if (operatorB.isDelimiter())
			return Operator.INTERACTION_COMMUTE_DELIMITER;
		Port portB2 = operatorB.getPort2();
		Port portC2 = portB2.getConnectedPort();
		portB2.disconnect();
		Delimiter delimiterDD =
			(Delimiter) this.cloneOperator(
				portB2.isInput() ^ this.hasOrientationReversed);
		Port portDD0 = delimiterDD.getPort0();
		Port portDD1 = delimiterDD.getPort1();
		portB2.connectTo(portDD1); // Not a principal connection.
		portDD0.connectTo(portC2);
		xListener.notifyConnection(portDD0, portC2);
		return operatorB.isAbstractor()
			? Operator.INTERACTION_EXTRUDE_ABSTRACTOR
			: operatorB.isApplicator()
			? Operator.INTERACTION_EXTRUDE_APPLICATOR
			: operatorB.isDuplicator()
			? Operator.INTERACTION_EXTRUDE_DUPLICATOR
			: 0;
	}

	/////////////////////////////////
	// IndexedOperator implementation
	/////////////////////////////////

	public int getIndex() {
		return this.index;
	}

	public void incrementIndex() {
		if (!this.hasOrientationReversed)
			this.index++;
	}

	///////////////////////////
	// Delimiter implementation
	///////////////////////////

	public Port getInput() {
		return this.input;
	}

	public Port getBody() {
		return this.body;
	}

	public void reverseOrientation() {
		if (this.hasOrientationReversed)
			throw new IllegalStateException();
		this.hasOrientationReversed = true;
		this.index = 0;
		this.input.number = 0;
		this.body.number = 1;
	}

	public ScopeBinder getScopeBinder() {
		return this.scopeBinder;
	}
}
