/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.impl;

import org.xmloperator.lambda.net.exception.AuxiliaryPortException;
import org.xmloperator.lambda.net.exception.MarkedDifferentlyException;
import org.xmloperator.lambda.net.model.ConnectionListener;
import org.xmloperator.lambda.net.model.Duplicator;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.Port;

/**
 * Duplicator implementation.
 */
final class DuplicatorImpl extends OperatorImpl implements Duplicator {
	final PortImpl output;
	final PortImpl input1;
	final PortImpl input2;
	int index = 0;

	DuplicatorImpl(
		OperatorFactoryImpl operatorFactory,
		int mark,
		boolean isDirectionInversed) {
		super(operatorFactory, mark);
		this.output =
			new PortImpl(Port.TYPE_DUPLICATOR_OUTPUT, this, 0, isDirectionInversed);
		this.input1 =
			new PortImpl(Port.TYPE_DUPLICATOR_INPUT1, this, 1, !isDirectionInversed);
		this.input2 =
			new PortImpl(Port.TYPE_DUPLICATOR_INPUT2, this, 2, !isDirectionInversed);
	}

	void setDirectionInversed(boolean isDirectionInversed) {
		this.output.isInput = isDirectionInversed;
		this.input1.isInput = !isDirectionInversed;
		this.input2.isInput = !isDirectionInversed;
	}

	//////////////////////////
	// Operator implementation
	//////////////////////////

	public boolean canBeReturnedToFactory() {
		return super.canBeReturnedToFactory()
			&& this.input1.connectedPort == null
			&& this.input2.connectedPort == null
			&& this.output.connectedPort == null;
	}

	public void returnToFactory() {
		super.returnToFactory();
		this.index = 0;
		this.operatorFactory.returnDuplicator(this);
	}

	public int getPortCount() {
		return 3;
	}

	public int getOperatorType() {
		return Operator.TYPE_DUPLICATOR;
	}

	public boolean isDuplicator() {
		return true;
	}

	public Port getPort(int number) {
		switch (number) {
			case 0 :
				return this.output;
			case 1 :
				return this.input1;
			case 2 :
				return this.input2;
			default :
				throw new IllegalArgumentException();
		}
	}

	public Port getPort0() {
		return this.output;
	}

	public Port getPort1() {
		return this.input1;
	}

	public Port getPort2() {
		return this.input2;
	}


	public boolean isDirectionInversed() {
		return this.output.isInput;
	}

	public Operator cloneOperator(boolean isDirectionInversed) {
		DuplicatorImpl duplicator =
			(DuplicatorImpl) this.operatorFactory.createDuplicator(
				this.mark,
				isDirectionInversed);
		duplicator.index = this.index;
		return duplicator;
	}

	/**
	 * The Duplicator "D" commutes with the operator "B" :
	 * (A--D>--<B--C) -> (A--<B--D>--C)
	 * or, if identical, annihilates it :
	 * (A--D>--<D--C) -> (A--C)
	 */
	public int interact(ConnectionListener xListener) {
		Port portD0 = this.getPort0();
		Port portB0 = portD0.getConnectedPort();
		Operator operatorB = portB0.getOperator();
		if (operatorB.isEraser() || operatorB.isDelimiter())
			return operatorB.interact(xListener);
		if (!portB0.isPrincipal())
			throw new AuxiliaryPortException(portB0);
		if (operatorB.getMark() != this.mark)
			throw new MarkedDifferentlyException(operatorB);
		Port portD1 = this.getPort1();
		Port portD2 = this.getPort2();
		Port portA1 = portD1.getConnectedPort();
		Port portA2 = portD2.getConnectedPort();
		portD0.disconnect();
		portD1.disconnect();
		portD2.disconnect();
		Port portB1 = operatorB.getPort1();
		Port portB2 = operatorB.getPort2();
		Port portC1 = portB1.getConnectedPort();
		Port portC2 = portB2.getConnectedPort();
		portB1.disconnect();
		portB2.disconnect();
		if (operatorB.isDuplicator()
			&& ((DuplicatorImpl) operatorB).index == this.index) {
			// Annihilates.
			portA1.connectTo(portC1);
			xListener.notifyConnection(portA1, portC1);
			portA2.connectTo(portC2);
			xListener.notifyConnection(portA2, portC2);
			operatorB.returnToFactory();
			this.returnToFactory();
			return Operator.INTERACTION_ANNIHILATE_DUPLICATOR;
		}
		// Commutes.
		Operator operatorBB =
			operatorB.cloneOperator(operatorB.isDirectionInversed());
		Port portBB0 = operatorBB.getPort0();
		Port portBB1 = operatorBB.getPort1();
		Port portBB2 = operatorBB.getPort2();
		portA1.connectTo(portB0);
		xListener.notifyConnection(portA1, portB0);
		portA2.connectTo(portBB0);
		xListener.notifyConnection(portA2, portBB0);
		if (portC1 == portB2) {
			if (portC2 != portB1 || !operatorB.isAbstractor())
				throw new IllegalStateException();
			// Operator B is an identity Abstractor
			portB1.connectTo(portB2); // Not a principal connection.
			portBB1.connectTo(portBB2); // Not a principal connection.
			this.returnToFactory();
			return Operator.INTERACTION_DUPLICATE_IDENTITY;
		}
		if (operatorB.isAbstractor())
			this.incrementIndex();
		DuplicatorImpl duplicatorDD =
			(DuplicatorImpl) this.cloneOperator(portB2.isInput());
		Port portDD0 = duplicatorDD.output;
		Port portDD1 = duplicatorDD.input1;
		Port portDD2 = duplicatorDD.input2;
		portD0.connectTo(portC1);
		xListener.notifyConnection(portD0, portC1);
		portDD0.connectTo(portC2);
		xListener.notifyConnection(portDD0, portC2);
		portB1.connectTo(portD1); // Not a principal connection.
		portB2.connectTo(portDD1); // Not a principal connection.
		portBB1.connectTo(portD2); // Not a principal connection.
		portBB2.connectTo(portDD2); // Not a principal connection.
		return operatorB.isAbstractor()
			? Operator.INTERACTION_DUPLICATE_ABSTRACTOR
			: operatorB.isApplicator()
			? Operator.INTERACTION_DUPLICATE_APPLICATOR
			: operatorB.isDuplicator()
			? Operator.INTERACTION_COMMUTE_DUPLICATOR
			: 0;
	}

	/////////////////////////////////
	// IndexedOperator implementation
	/////////////////////////////////

	public int getIndex() {
		return this.index;
	}

	public void incrementIndex() {
		this.index++;
	}

	////////////////////////////
	// Duplicator implementation
	////////////////////////////

	public Port getInput1() {
		return this.input1;
	}

	public Port getInput2() {
		return this.input2;
	}

	public Port getOutput() {
		return this.output;
	}
}
