/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.impl;

import org.xmloperator.lambda.net.exception.AuxiliaryPortException;
import org.xmloperator.lambda.net.exception.MarkedDifferentlyException;
import org.xmloperator.lambda.net.model.Abstractor;
import org.xmloperator.lambda.net.model.ConnectionListener;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;

/**
 * Eraser implementation.
 */
final class EraserImpl extends OperatorImpl implements Eraser {
	final PortImpl port;

	EraserImpl(
		OperatorFactoryImpl operatorFactory,
		int mark,
		boolean isDirectionInversed,
		boolean isActive) {
		super(operatorFactory, mark);
		this.port =
			new PortImpl(
				Port.TYPE_ERASER_PORT,
				this,
				isActive ? 0 : 1,
				isDirectionInversed);
	}

	void setDirectionInversed(boolean isDirectionInversed) {
		this.port.isInput = isDirectionInversed;
	}

	public void setActive(boolean isActive) {
		this.port.number = isActive ? 0 : 1;
	}

	//////////////////////////
	// Operator implementation
	//////////////////////////

	public boolean canBeReturnedToFactory() {
		return super.canBeReturnedToFactory() && this.port.connectedPort == null;
	}

	public void returnToFactory() {
		super.returnToFactory();
		this.port.number = 0;
		this.operatorFactory.returnEraser(this);
	}

	public int getPortCount() {
		return 1;
	}

	public int getOperatorType() {
		return Operator.TYPE_ERASER;
	}

	public boolean isEraser() {
		return true;
	}

	public Port getPort(int number) {
		if (this.port.number == number)
			return this.port;
		if (number >= 0 && number <= 2)
			return null;
		throw new IllegalArgumentException();
	}

	public Port getPort0() {
		return this.port.number == 0 ? this.port : null;
	}

	public Port getPort1() {
		return this.port.number == 1 ? this.port : null;
	}

	public Port getPort2() {
		return null;
	}

	public boolean isDirectionInversed() {
		return this.port.isInput;
	}

	public Operator cloneOperator(boolean isDirectionInversed) {
		return this.operatorFactory.createEraser(
			this.mark,
			isDirectionInversed,
			this.port.number == 0);
	}

	/**
	 * The Eraser "E" erases the Operator "A" :
	 * (E>--<A--B) -> (E>--B)
	 */
	public int interact(ConnectionListener xListener) {
		if (this.port.number != 0)
			throw new IllegalStateException();
		PortImpl portE1 = this.port;
		PortImpl portA0 = portE1.connectedPort;
		OperatorImpl operatorA = portA0.operator;
		if (!portA0.isPrincipal())
			throw new AuxiliaryPortException(portA0);
		int mark = this.mark;
		if (operatorA.getMark() != mark)
			throw new MarkedDifferentlyException(operatorA);
		portE1.disconnect();
		switch (operatorA.getPortCount()) {
			case 1 :
				operatorA.returnToFactory();
				this.returnToFactory();
				return Operator.INTERACTION_ANNIHILATE_ERASER;
			case 2 :
				{
					Port portB1 = operatorA.getPort1().getConnectedPort();
					portB1.disconnect();
					portE1.connectTo(portB1);
					xListener.notifyConnection(portE1, portB1);
					operatorA.returnToFactory();
					return Operator.INTERACTION_ERASE_DELIMITER;
				}
			case 3 :
				{
					if (operatorA.isAbstractor()
						&& ((Abstractor) operatorA).isIdentity()) {
						operatorA.getPort1().disconnect();
						operatorA.returnToFactory();
						this.returnToFactory();
						return Operator.INTERACTION_ERASE_IDENTITY;
					}
					Port portA1 = operatorA.getPort1();
					Port portA2 = operatorA.getPort2();
					Port portB1 = portA1.getConnectedPort();
					Port portB2 = portA2.getConnectedPort();
					portA1.disconnect();
					portA2.disconnect();
					if (portB1 == portA2 || portB2 == portA1)
						throw new IllegalStateException();
					OperatorFactory operatorFactory = this.operatorFactory;
					portE1.connectTo(portB1);
					xListener.notifyConnection(portE1, portB1);
					Port portE2 =
						operatorFactory
							.createEraser(mark, !portB2.isInput(), true)
							.getPort0();
					portE2.connectTo(portB2);
					xListener.notifyConnection(portE2, portB2);
					int interactionType =
						operatorA.isDelimiter()
							? Operator.INTERACTION_ERASE_DELIMITER
							: operatorA.isAbstractor()
							? Operator.INTERACTION_ERASE_ABSTRACTOR
							: operatorA.isApplicator()
							? Operator.INTERACTION_ERASE_APPLICATOR
							: operatorA.isDuplicator()
							? Operator.INTERACTION_ERASE_DUPLICATOR
							: 0;
					operatorA.returnToFactory();
					return interactionType;
				}
			default :
				throw new IllegalStateException();
		}
	}

	////////////////////////
	// Eraser implementation
	////////////////////////

	public Port getPort() {
		return this.port;
	}
}
