/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.impl;

import org.xmloperator.lambda.net.model.Abstractor;
import org.xmloperator.lambda.net.model.Applicator;
import org.xmloperator.lambda.net.model.Delimiter;
import org.xmloperator.lambda.net.model.Duplicator;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.PortStack;
import org.xmloperator.lambda.net.model.ScopeBinder;
import org.xmloperator.lambda.net.model.Walk;
import org.xmloperator.lambda.net.model.WalkStack;
import org.xmloperator.lambda.shared.ObjectArrayFactory;
import org.xmloperator.lambda.shared.Stack;

/**
 * OperatorFactory implementation.
 * Can either reuse objects or garbage them.
 */
public class OperatorFactoryImpl implements OperatorFactory {
	private final boolean areObjectsReused;
	private final Stack eraserStack = new Stack();
	private final Stack delimiterStack = new Stack();
	private final Stack abstractorStack = new Stack();
	private final Stack applicatorStack = new Stack();
	private final Stack duplicatorStack = new Stack();
	private final ObjectArrayFactory objectArrayFactory;
	private final Stack portStackStack = new Stack();
	private final Stack scopeBinderStack = new Stack();
	private final Stack directorStepStack = new Stack();
	private final Stack walkStack = new Stack();
	private final Stack walkStackStack = new Stack();
	private int newEraserCount = 0;
	private int newDelimiterCount = 0;
	private int newAbstractorCount = 0;
	private int newApplicatorCount = 0;
	private int newDuplicatorCount = 0;
	private int newPortStackCount = 0;
	private int newScopeBinderCount = 0;
	private int newDirectorStepCount = 0;
	private int newWalkCount = 0;
	private int newWalkStackCount = 0;

	/**
	 * OperatorFactory constructor.
	 * @param areObjectsReused:
	 * 	 - true: returned objects are reused.
	 *   - false: returned objects are garbaged.
	 * @param objectArrayFactory an ObjectArrayFactory.
	 */
	public OperatorFactoryImpl(
		boolean areObjectsReused,
		ObjectArrayFactory objectArrayFactory) {
		this.areObjectsReused = areObjectsReused;
		this.objectArrayFactory = objectArrayFactory;
	}

	/**
	 * Clear memory.
	 */
	public void clear() {
		this.eraserStack.clear();
		this.delimiterStack.clear();
		this.abstractorStack.clear();
		this.applicatorStack.clear();
		this.duplicatorStack.clear();
		this.portStackStack.clear();
		this.scopeBinderStack.clear();
		this.directorStepStack.clear();
		this.walkStack.clear();
		this.walkStackStack.clear();
		this.newEraserCount = 0;
		this.newDelimiterCount = 0;
		this.newAbstractorCount = 0;
		this.newApplicatorCount = 0;
		this.newDuplicatorCount = 0;
		this.newPortStackCount = 0;
		this.newScopeBinderCount = 0;
		this.newDirectorStepCount = 0;
		this.newWalkCount = 0;
		this.newWalkStackCount = 0;
	}

	/**
	 * @return true if all the objects that have been created have been returned.
	 */
	public boolean isWellBalanced() {
		return this.eraserStack.getSize() == this.newEraserCount
			&& this.delimiterStack.getSize() == this.newDelimiterCount
			&& this.abstractorStack.getSize() == this.newAbstractorCount
			&& this.applicatorStack.getSize() == this.newApplicatorCount
			&& this.duplicatorStack.getSize() == this.newDuplicatorCount
			&& this.portStackStack.getSize() == this.newPortStackCount
			&& this.scopeBinderStack.getSize() == this.newScopeBinderCount
			&& this.directorStepStack.getSize() == this.newDirectorStepCount
			&& this.walkStack.getSize() == this.newWalkCount
			&& this.walkStackStack.getSize() == this.newWalkStackCount;
	}

	public String toString() {
		return "[Eraser("
			+ this.newEraserCount
			+ ","
			+ this.eraserStack.getSize()
			+ "), Delimiter("
			+ this.newDelimiterCount
			+ ","
			+ this.delimiterStack.getSize()
			+ "), Abstractor("
			+ this.newAbstractorCount
			+ ","
			+ this.abstractorStack.getSize()
			+ "), Applicator("
			+ this.newApplicatorCount
			+ ","
			+ this.applicatorStack.getSize()
			+ "), Duplicator("
			+ this.newDuplicatorCount
			+ ","
			+ this.duplicatorStack.getSize()
			+ "), PortStack("
			+ this.newPortStackCount
			+ ","
			+ this.portStackStack.getSize()
			+ "), ScopeBinder("
			+ this.newScopeBinderCount
			+ ","
			+ this.scopeBinderStack.getSize()
			+ "), DirectorStep("
			+ this.newDirectorStepCount
			+ ","
			+ this.directorStepStack.getSize()
			+ "), Walk("
			+ this.newWalkCount
			+ ","
			+ this.walkStack.getSize()
			+ "), WalkStack("
			+ this.newWalkStackCount
			+ ","
			+ this.walkStackStack.getSize()
			+ ")]";
	}

	void returnEraser(Eraser eraser) {
		if (this.areObjectsReused)
			this.eraserStack.push(eraser);
	}

	void returnDelimiter(Delimiter delimiter) {
		if (this.areObjectsReused)
			this.delimiterStack.push(delimiter);
	}

	void returnAbstractor(Abstractor abstractor) {
		if (this.areObjectsReused)
			this.abstractorStack.push(abstractor);
	}

	void returnApplicator(Applicator applicator) {
		if (this.areObjectsReused)
			this.applicatorStack.push(applicator);
	}

	void returnDuplicator(Duplicator duplicator) {
		if (this.areObjectsReused)
			this.duplicatorStack.push(duplicator);
	}

	void returnPortStack(PortStack portStack) {
		if (this.areObjectsReused)
			this.portStackStack.push(portStack);
	}

	void returnScopeBinder(ScopeBinder scopeBinder) {
		if (this.areObjectsReused)
			this.scopeBinderStack.push(scopeBinder);
	}

	DirectorStep createDirectorStep(int director, DirectorStep nextStep) {
		if (this.areObjectsReused) {
			DirectorStep directorStep = (DirectorStep) this.directorStepStack.pop();
			if (directorStep != null) {
				directorStep.director = director;
				directorStep.nextStep = nextStep;
				return directorStep;
			}
		}
		this.newDirectorStepCount++;
		return new DirectorStep(director, nextStep);
	}

	void returnDirectorStep(DirectorStep directorStep) {
		directorStep.nextStep = null;
		if (this.areObjectsReused)
			this.directorStepStack.push(directorStep);
	}

	void returnWalk(Walk walk) {
		if (this.areObjectsReused)
			this.walkStack.push(walk);
	}

	void returnWalkStack(WalkStack walkStack) {
		if (this.areObjectsReused)
			this.walkStackStack.push(walkStack);
	}

	/////////////////////////////////
	// OperatorFactory implementation
	/////////////////////////////////

	public Eraser createEraser(
		int mark,
		boolean isDirectionInversed,
		boolean isActive) {
		if (this.areObjectsReused) {
			EraserImpl eraser = (EraserImpl) this.eraserStack.pop();
			if (eraser != null) {
				eraser.isReturnedToTheFactory = false;
				eraser.mark = mark;
				eraser.setDirectionInversed(isDirectionInversed);
				eraser.setActive(isActive);
				return eraser;
			}
		}
		this.newEraserCount++;
		return new EraserImpl(this, mark, isDirectionInversed, isActive);
	}

	public Delimiter createDelimiter(
		int mark,
		boolean isDirectionInversed,
		ScopeBinder scopeBinder) {
		if (this.areObjectsReused) {
			DelimiterImpl delimiter = (DelimiterImpl) this.delimiterStack.pop();
			if (delimiter != null) {
				delimiter.isReturnedToTheFactory = false;
				delimiter.mark = mark;
				delimiter.setDirectionInversed(isDirectionInversed);
				delimiter.scopeBinder = scopeBinder;
				if (scopeBinder != null)
					scopeBinder.getDelimiterList().add(delimiter);
				return delimiter;
			}
		}
		this.newDelimiterCount++;
		DelimiterImpl delimiter =
			new DelimiterImpl(this, mark, isDirectionInversed);
		delimiter.scopeBinder = scopeBinder;
		if (scopeBinder != null)
			scopeBinder.getDelimiterList().add(delimiter);
		return delimiter;
	}

	public Abstractor createAbstractor(int mark, ScopeBinder scopeBinder) {
		if (this.areObjectsReused) {
			AbstractorImpl abstractor = (AbstractorImpl) this.abstractorStack.pop();
			if (abstractor != null) {
				abstractor.isReturnedToTheFactory = false;
				abstractor.mark = mark;
				abstractor.scopeBinder = scopeBinder;
				if (scopeBinder != null)
					scopeBinder.getAbstractorList().add(abstractor);
				return abstractor;
			}
		}
		this.newAbstractorCount++;
		AbstractorImpl abstractor =
			new AbstractorImpl(this, this.objectArrayFactory, mark);
		abstractor.scopeBinder = scopeBinder;
		if (scopeBinder != null)
			scopeBinder.getAbstractorList().add(abstractor);
		return abstractor;
	}

	public Applicator createApplicator(int mark) {
		if (this.areObjectsReused) {
			ApplicatorImpl applicator = (ApplicatorImpl) this.applicatorStack.pop();
			if (applicator != null) {
				applicator.isReturnedToTheFactory = false;
				applicator.mark = mark;
				return applicator;
			}
		}
		this.newApplicatorCount++;
		return new ApplicatorImpl(this, mark);
	}

	public Duplicator createDuplicator(int mark, boolean isDirectionInversed) {
		if (this.areObjectsReused) {
			DuplicatorImpl duplicator = (DuplicatorImpl) this.duplicatorStack.pop();
			if (duplicator != null) {
				duplicator.isReturnedToTheFactory = false;
				duplicator.mark = mark;
				duplicator.setDirectionInversed(isDirectionInversed);
				return duplicator;
			}
		}
		this.newDuplicatorCount++;
		return new DuplicatorImpl(this, mark, isDirectionInversed);
	}

	public PortStack createPortStack() {
		if (this.areObjectsReused) {
			PortStackImpl portStack = (PortStackImpl) this.portStackStack.pop();
			if (portStack != null) {
				portStack.startUse();
				return portStack;
			}
		}
		this.newPortStackCount++;
		PortStackImpl portStack = new PortStackImpl(this, this.objectArrayFactory);
		portStack.startUse();
		return portStack;
	}

	public ScopeBinder createScopeBinder() {
		if (this.areObjectsReused) {
			ScopeBinderImpl scopeBinder =
				(ScopeBinderImpl) this.scopeBinderStack.pop();
			if (scopeBinder != null) {
				scopeBinder.isReturnedToTheFactory = false;
				scopeBinder.startUse();
				return scopeBinder;
			}
		}
		this.newScopeBinderCount++;
		ScopeBinderImpl scopeBinder =
			new ScopeBinderImpl(this, this.objectArrayFactory);
		scopeBinder.startUse();
		return scopeBinder;
	}

	public Walk createWalk(int bindIndex) {
		if (this.areObjectsReused) {
			WalkImpl walk = (WalkImpl) this.walkStack.pop();
			if (walk != null) {
				walk.isReturnedToTheFactory = false;
				walk.bindIndex = bindIndex;
				return walk;
			}
		}
		this.newWalkCount++;
		return new WalkImpl(this, bindIndex);
	}

	public WalkStack createWalkStack() {
		if (this.areObjectsReused) {
			WalkStackImpl walkStack = (WalkStackImpl) this.walkStackStack.pop();
			if (walkStack != null) {
				walkStack.startUse();
				return walkStack;
			}
		}
		this.newWalkStackCount++;
		WalkStackImpl walkStack = new WalkStackImpl(this, this.objectArrayFactory);
		walkStack.startUse();
		return walkStack;
	}
}
