/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.model;

/**
 * Operator node in a lambda-net.
 * <p>It is either an <tt>Eraser</tt>, a <tt>Delimiter</tt>, an <tt>Abstractor</tt>,
 *   an <tt>Applicator</tt> or a <tt>Duplicator</tt>.</p>
 * <p>Each one has a principal Port and may have 1 or 2 auxiliary Ports.
 *   By exception, the Eraser has either a principal Port either an auxiliary
 *   Port. The principal Port is numbered 0. Auxiliary ports are numbered 1 and
 *   2.</p>
 * <p>An Operator instance is created by using an <tt>OperatorFactory</tt> and
 *   is terminated by being returned to the same <tt>OperatorFactory</tt>.</p>
 * @see org.xmloperator.lambda.net.model.OperatorFactory
 */
public interface Operator {

	/**
	 * Returns the factory of this Operator, which can be used for building
	 *   compatible Operators.
	 * @return the factory of this Operator.
	 */
	public OperatorFactory getFactory();

	/**
	 * Returns <tt>true</tt> if this Operator has been returned to the factory.
	 * <p>It must be avoided to keep a reference to an Operator that has been
	 *   returned to the factory.</p>
	 * <p>This method is the only one that is allowed to call if this Operator has
	 *   been returned to the factory.</p>
	 * @return <tt>true</tt> if this Operator has been returned to the factory.
	 */
	public boolean isReturnedToFactory();

	/**
	 * Returns <tt>true</tt> if this can be returned to its factory.
	 * <p>If this is connected then this cannot be returned to factory.
	 * It may exist other required conditions for being returnable.</p>
	 * @return <tt>true</tt> if this can be returned to the factory.
	 */
	public boolean canBeReturnedToFactory();

	/**
	 * Returns this to its Factory.
	 * <p><b>Pre-condition:</b> this can be returned to the factory.</p>
	 * <p><b>Post-condition:</b> this is no more usable.</p>
	 * @throws IllegalStateException if this cannot be returned to the factory.
	 */
	public void returnToFactory();

	/**
	 * Returns the count of ports of this Operator.
	 * This is <tt>1</tt> for an <tt>Eraser</tt>,
	 *   <tt>2</tt> for a <tt>Delimiter</tt>,
	 *   <tt>3</tt> for an <tt>Applicator</tt>, a <tt>Duplicator</tt> or an
	 *   <tt>Abstractor</tt>.
	 * @return the count of ports of this Operator.
	 */
	public int getPortCount();

	public static final int OPERATOR_TYPE_COUNT = 5;
	public static final int TYPE_ERASER = 0;
	public static final int TYPE_DELIMITER = 1;
	public static final int TYPE_ABSTRACTOR = 2;
	public static final int TYPE_APPLICATOR = 3;
	public static final int TYPE_DUPLICATOR = 4;
	public static final String[] OPERATOR_TYPE_NAMES =
		{ "Eraser", "Delimiter", "Abstractor", "Applicator", "Duplicator" };

	/**
	 * Returns the type of this Operator. It is one of the constants TYPE_*.
	 * @return the type of this Operator.
	 */
	public int getOperatorType();

	/**
	 * Returns <tt>true</tt> if this Operator is an Eraser.
	 * @return <tt>true</tt> if this Operator is an Eraser.
	 */
	public boolean isEraser();

	/**
	 * Returns <tt>true</tt> if this Operator is a Delimiter.
	 * @return <tt>true</tt> if this Operator is a Delimiter.
	 */
	public boolean isDelimiter();

	/**
	 * Returns <tt>true</tt> if this Operator is an Abstractor.
	 * @return <tt>true</tt> if this Operator is an Abstractor.
	 */
	public boolean isAbstractor();

	/**
	 * Returns <tt>true</tt> if this Operator is an Applicator.
	 * @return <tt>true</tt> if this Operator is an Applicator.
	 */
	public boolean isApplicator();

	/**
	 * Returns <tt>true</tt> if this Operator is a Duplicator.
	 * @return <tt>true</tt> if this Operator is a Duplicator.
	 */
	public boolean isDuplicator();

	/**
	 * Returns the Port of this Operator, with the given number. It may depend on
	 *   the Operator orientation. Return <tt>null</tt> if it doesn't exist.
	 * @return the Port of this Operator, with the given number.
	 * @param number the Port number = 0, 1 or 2.
	 * @throws IllegalArgumentException if the number is neither 0, 1 nor 2.
	 */
	public Port getPort(int number);

	/**
	 * Returns the principal Port of this Operator or null if it doesn't exist.
	 * This is a shortcut for getPort(0).
	 * @return the principal Port of this Operator or null if it doesn't exist.
	 */
	public Port getPort0();

	/**
	 * Returns the first auxiliary Port of this Operator or null if it doesn't
	 *   exist. This is a shortcut for getPort(1).
	 * @return the first auxiliary Port of this Operator or null if it doesn't
	 *   exist.
	 */
	public Port getPort1();

	/**
	 * Returns the second auxiliary Port of this Operator or null if it doesn't
	 *   exist. This is a shortcut for getPort(2).
	 * @return the second auxiliary Port of this Operator or null if it doesn't
	 *   exist.
	 */
	public Port getPort2();

	/**
	 * Operators can be marked for net traversal. Returns the mark attached to
	 *   this Operator.
	 * @return the mark of this Operator.
	 */
	public int getMark();

	/**
	 * Set the mark attached to this operator.
	 * @param mark a mark.
	 */
	public void setMark(int mark);

	/**
	 * Returns <tt>true</tt> if the direction of this Operator is inversed.
	 * @return <tt>true</tt> if the direction of this Operator is inversed.
	 */
	public boolean isDirectionInversed();

	/**
	 * Creates and return a clone of this Operator.
	 * Doesn't take current connections into account. The clone is disconnected.
	 * Takes the mark into account. The clone will have the same mark.
	 * @param isDirectionInversed if <tt>true</tt> then the the clone Operator
	 *   direction will be inversed (independently from the ditection of this
	 *   Operator).
	 * @return the created clone of this Operator.
	 */
	public Operator cloneOperator(boolean isDirectionInversed);

	public static final int INTERACTION_TYPE_COUNT = 20;
	public static final int INTERACTION_BETA_REDUCTION = 0;
	public static final int INTERACTION_BETA_IDENTITY = 1;
	public static final int INTERACTION_EXTRUDE_ABSTRACTOR = 2;
	public static final int INTERACTION_EXTRUDE_IDENTITY = 3;
	public static final int INTERACTION_EXTRUDE_CLOSED_ABSTRACTOR = 4;
	public static final int INTERACTION_EXTRUDE_APPLICATOR = 5;
	public static final int INTERACTION_EXTRUDE_DUPLICATOR = 6;
	public static final int INTERACTION_ANNIHILATE_ERASER = 7;
	public static final int INTERACTION_ANNIHILATE_DELIMITER = 8;
	public static final int INTERACTION_ANNIHILATE_DUPLICATOR = 9;
	public static final int INTERACTION_COMMUTE_DELIMITER = 10;
	public static final int INTERACTION_COMMUTE_DUPLICATOR = 11;
	public static final int INTERACTION_DUPLICATE_ABSTRACTOR = 12;
	public static final int INTERACTION_DUPLICATE_IDENTITY = 13;
	public static final int INTERACTION_DUPLICATE_APPLICATOR = 14;
	public static final int INTERACTION_ERASE_DELIMITER = 15;
	public static final int INTERACTION_ERASE_ABSTRACTOR = 16;
	public static final int INTERACTION_ERASE_IDENTITY = 17;
	public static final int INTERACTION_ERASE_APPLICATOR = 18;
	public static final int INTERACTION_ERASE_DUPLICATOR = 19;
	public static final String[] INTERACTION_TYPE_NAMES =
		{
			"BetaReduction",
			"BetaIdentity",
			"ExtrudeAbstractor",
			"ExtrudeIdentity",
			"ExtrudeClosedAbstractor",
			"ExtrudeApplicator",
			"ExtrudeDuplicator",
			"AnnihilateEraser",
			"AnnihilateDelimiter",
			"AnnihilateDuplicator",
			"CommuteDelimiter",
			"CommuteDuplicator",
			"DuplicateAbstractor",
			"DuplicateIdentity",
			"DuplicateApplicator",
			"EraseDelimiter",
			"EraseAbstractor",
			"EraseIdentity",
			"EraseApplicator",
			"EraseDuplicator" };

	/**
	 * Interacts with the Operator connected on its principal Port.
	 * Notify any connection (except beta) between two principal Ports
	 *   through a ConnectionListener.
	 * Returns the conclusion of the interaction.
	 * <p><b>Pre-conditions:</b></p><ul>
	 * <li>the connected Port is a principal Port.</li>
	 * <li>the connected Operators have the same mark.</li>
	 * <li>the interaction net is closed.</li></ul>
	 * <p><b>Post-conditions:</b></p><ul>
	 * <li>the created Operators, if any, have the same mark.</li>
	 * <li>the interaction net is closed.</li></ul>
	 * @param xListener a ConnectionListener for notifying principal connections
	 *   other than beta.
	 * @return the report of the interaction : one of the INTERACTION_* constants.
	 * @throws NullPointerException if the principal Port is not connected or
	 *   if the interaction meets a free Port.
	 * @throws org.xmloperator.lambda.net.exception.AuxiliaryPortException
	 *   if the connected Port is not a principal Port.
	 * @throws org.xmloperator.lambda.net.exception.MarkedDifferentlyException
	 *   if a connected Operator has not the same mark.
	 * @throws org.xmloperator.lambda.net.exception.UnexpectedAnnihilationException
	 *   if the interacting Operators are two Abstractors or two Applicators. 
	 */
	public int interact(ConnectionListener xListener);
}
