/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.reduction;

import org.xmloperator.lambda.net.model.Applicator;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.model.Port;
import org.xmloperator.lambda.net.model.PortStack;
import org.xmloperator.lambda.net.util.InteractionListener;
import org.xmloperator.lambda.net.util.NetUtils;
import org.xmloperator.lambda.net.util.WalkingContext;

/**
 * Beta reductions in a lambda-net.
 */
public abstract class NetBetaUtils {
	private static final boolean ASSERTION = Test.ASSERTION;
	private static final boolean DEBUG = false;

	/**
	 * Returns <tt>true</tt> if the given <tt>Applicator</tt> is a redex.
	 * @param applicator an <tt>Applicator</tt>.
	 * @return <tt>true</tt> if the given <tt>Applicator</tt> is a redex.
	 */
	public static final boolean isRedex(Applicator applicator) {
		Port connectedPort = applicator.getBody().getConnectedPort();
		return connectedPort.isPrincipal()
			&& connectedPort.getOperator().isAbstractor();
	}

	/**
	 * Return the leftermost-outermost redex <tt>Applicator</tt> of a lambda-net
	 *   or null if the lambda-net has no redex.
	 * <p><b>Pre-conditions:</b></p><ul>
	 * <li>there is no reducable interaction other than beta and the root
	 *   <tt>Eraser</tt>.</li>
	 * <li>the <tt>WalkingContext</tt> is positioned to the root <tt>Eraser</tt>.</li></ul>
	 * <p><b>Post-condition:</b> the <tt>WalkingContext</tt> is positioned to the
	 *   root <tt>Eraser</tt>.</p>
	 * @param walkingContext a <tt>WalkingContext</tt>.
	 * @return the leftermost-outermost redex <tt>Applicator</tt> or null if none.
	 * @throws IllegalStateException if the given <tt>WalkingContext</tt> is not
	 *   curently positioned to the root <tt>Eraser</tt>.
	 * @throws NullPointerException if the given lambda-net is not closed.
	 */
	public static final Applicator leftermostOutermostRedex(final WalkingContext walkingContext) {
		if (!walkingContext.isReturnedToRoot())
			throw new IllegalStateException();
		if (ASSERTION)
			NetUtils.verifyXFree(walkingContext.getRootEraser());
		while (true) {
			walkingContext.executeStep();
			Port port = walkingContext.getCurrentPort();
			Operator operator = port.getOperator();
			if (operator.isApplicator()) {
				if (port.isPrincipal() && isRedex((Applicator) operator)) {
					walkingContext.returnToRoot();
					return (Applicator) operator;
				}
			}
			else if (operator.isEraser())
				return null;
		}
	}

	/**
	 * Reduce a beta interaction and apply the x-rules.
	 * <p><b>Pre-conditions:</b></p><ul>
	 * <li>there is no reducable interaction other than beta and the root
	 *   <tt>Eraser</tt>.</li>
	 * <li>the root <tt>Eraser</tt> is inactive.</li></ul>
	 * <p><b>Post-conditions:</b></p><ul>
	 * <li>there is no reducable interaction other than beta and the root
	 *   <tt>Eraser</tt>.</li>
	 * <li>the root <tt>Eraser</tt> is inactive.</li></ul>
	 * @param rootEraser a root <tt>Eraser</tt>.
	 * @param redexApplicator a redex <tt>Applicator</tt>.
	 * @param interactionListener an <tt>InteractionListener</tt>. May be null.
	 */
	public static final void reduce(
		final Eraser rootEraser,
		final Applicator redexApplicator,
		final InteractionListener interactionListener) {
		if (ASSERTION) {
			if (!isRedex(redexApplicator))
				throw new IllegalArgumentException();
			NetUtils.verifyXFree(rootEraser);
		}
		OperatorFactory operatorFactory = redexApplicator.getFactory();
		PortStack xPortStack = operatorFactory.createPortStack();
		try {
			{
				int interactionType = redexApplicator.interact(xPortStack);
				if (ASSERTION)
					NetUtils.verifyClosure(rootEraser);
				if (interactionListener != null)
					interactionListener.notifyInteraction(interactionType);
			}
			applyXRules(rootEraser, xPortStack, interactionListener);
		}
		finally {
			xPortStack.returnToFactory();
		}
		if (ASSERTION)
			NetUtils.verifyXFree(rootEraser);
	}

	private static final void applyXRules(
		final Eraser rootEraser,
		final PortStack xPortStack,
		final InteractionListener interactionListener) {
		while (!xPortStack.isEmpty()) {
			Operator operator = xPortStack.pop().getOperator();
			if (!operator.isReturnedToFactory()) {
				int interactionType = operator.interact(xPortStack);
				if (false)
					System.out.println(
						"interaction : "
							+ Operator.INTERACTION_TYPE_NAMES[interactionType]);
				if (ASSERTION)
					NetUtils.verifyClosure(rootEraser);
				if (interactionListener != null)
					interactionListener.notifyInteraction(interactionType);
			}
		}
	}

	/**
	 * Reduce a lambda-net until a normal form is obtained or an exception is
	 *   throwed.
	 * <p><b>Pre-conditions:</b></p><ul>
	 * <li>there is no reducable interaction other than beta and the root
	 *   <tt>Eraser</tt>.</li>
	 * <li>the root <tt>Eraser</tt> is inactive.</li>
	 * <li>the <tt>WalkingContext</tt> is positioned to the root <tt>Eraser</tt>.</li></ul>
	 * <p><b>Post-conditions:</b></p><ul>
	 * <li>there is no reducable interaction other than the root <tt>Eraser</tt>.</li>
	 * <li>the root <tt>Eraser</tt> is inactive.</li>
	 * <li>the <tt>WalkingContext</tt> is positioned to the root <tt>Eraser</tt>.</li></ul>
	 * @param walkingContext a <tt>WalkingContext</tt>.
	 * @param interactionListener an <tt>InteractionListener</tt>. May be null.
	 * @param redexListener a <tt>RedexListener</tt>. May be null.
	 * @throws NullPointerException if the lambda-net is broken.
	 * @throws NotInNormalFormException if the <tt>ReductionListener</tt> decides
	 *   to throw this exception in order to stop this process.
	 */
	public static final void reduceDeeply(
		WalkingContext walkingContext,
		InteractionListener interactionListener,
		ReductionListener redexListener) {
		if (!walkingContext.isReturnedToRoot())
			throw new IllegalStateException();
		Eraser rootEraser = walkingContext.getRootEraser();
		while (true) {
			if (DEBUG) {
				NetUtils.print(System.out, rootEraser);
				walkingContext.printTour(System.out);
			}
			Applicator redexApplicator = leftermostOutermostRedex(walkingContext);
			if (redexApplicator == null)
				break;
			if (redexListener != null)
				redexListener.notifyStartReduction(rootEraser, redexApplicator);
			reduce(rootEraser, redexApplicator, interactionListener);
		}
		if (!walkingContext.isReturnedToRoot()
			|| (ASSERTION && leftermostOutermostRedex(walkingContext) != null))
			throw new IllegalStateException();
	}
}
