/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.util;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.lambda.net.exception.UnreducedXInteractionException;
import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.IndexedOperator;
import org.xmloperator.lambda.net.model.Operator;
import org.xmloperator.lambda.net.model.Port;

/**
 * Utilities about traversing a lambda-net.
 */
public abstract class NetUtils {

	/**
	 * Increment (or decrement) the mark of any <tt>Operator</tt> connected to
	 *   an <tt>Eraser</tt>, recursively.
	 * <p><b>Pre-condition:</b> all the <tt>Operator</tt>s have the same mark.</p>
	 * <p><b>Post-condition:</b> the <tt>Operator</tt> marks are incremented.</p>
	 * @param rootEraser a root <tt>Eraser</tt>.
	 * @throws FreePortException if a <tt>Port</tt> is found free.
	 * @throws MarkedDifferentlyException if an <tt>Operator</tt> has not
	 *   the same mark than the others.
	 */
	public static final void incrementMark(
		Eraser rootEraser,
		boolean isMarkIncremented) {
		OperatorIterator iterator = new NetTraversal(rootEraser, isMarkIncremented);
		while (iterator.hasNext())
			iterator.next();
	}

	/**
	 * Verify the closure and the mark-homogeneity of a lambda-net.
	 * <p><b>Pre-condition:</b> all the <tt>Operator</tt>s have the same mark.</p>
	 * <p><b>Post-condition:</b> the <tt>Operator</tt> marks are unmodified.</p>
	 * @param rootEraser a root <tt>Eraser</tt>.
	 * @throws FreePortException if a <tt>Port</tt> is found free.
	 * @throws MarkedDifferentlyException if an <tt>Operator</tt> has not
	 *   the same mark than the others.
	 */
	public static final void verifyClosure(Eraser rootEraser) {
		incrementMark(rootEraser, true);
		incrementMark(rootEraser, false);
	}

	/**
	 * Verify that there is no reducable interaction other than beta and the
	 *   root Eraser.
	 * <p><b>Pre-condition:</b> all the <tt>Operator</tt>s have the same mark.</p>
	 * <p><b>Post-condition:</b> the <tt>Operator</tt> marks are unmodified.</p>
	 * @param rootEraser a root <tt>Eraser</tt>.
	 * @throws UnreducedXInteractionException if there is a reducable
	 *   interaction other than beta and the root <tt>Eraser</tt>.
	 * @throws FreePortException if a <tt>Port</tt> is found free.
	 * @throws MarkedDifferentlyException if an <tt>Operator</tt> has not
	 *   the same mark than the others.
	 */
	public static final void verifyXFree(Eraser rootEraser) {
		OperatorIterator iterator = new NetTraversal(rootEraser, true);
		iterator.next(); // Root Eraser.
		while (iterator.hasNext()) {
			Operator operator = iterator.next();
			Port connectedPort = operator.getPort0().getConnectedPort();
			if (connectedPort.isPrincipal()) {
				Operator interactingOperator = connectedPort.getOperator();
				if (!((operator.isApplicator() && interactingOperator.isAbstractor())
					|| (operator.isAbstractor() && interactingOperator.isApplicator())))
					throw new UnreducedXInteractionException(connectedPort);
			}
		}
		incrementMark(rootEraser, false);
	}

	/**
	 * Returns a List of <tt>Operator</tt>s belonging to a lambda-net.
	 * @param rootEraser a root <tt>Eraser</tt>.
	 * @return a List of the <tt>Operator</tt>s of the given interaction net.
	 * @throws FreePortException if a <tt>Port</tt> is found free.
	 * @throws MarkedDifferentlyException if an <tt>Operator</tt> has not
	 *   the same mark than the others.
	 */
	public static final List operatorList(Eraser rootEraser) {
		List operatorList = new ArrayList();
		OperatorIterator iterator = new NetTraversal(rootEraser, true);
		while (iterator.hasNext())
			operatorList.add(iterator.next());
		incrementMark(rootEraser, false);
		return operatorList;
	}

	/**
	 * Print a presentation of a lambda-net.
	 * @param out a <tt>PrintStream</tt>.
	 * @param rootEraser a root <tt>Eraser</tt>.
	 */
	public static final void print(PrintStream out, Eraser rootEraser) {
		print(out, operatorList(rootEraser));
	}

	/**
	 * Print a presentation of a lambda-net.
	 * @param out a <tt>PrintStream</tt>.
	 * @param operatorList a List of <tt>Operator</tt>s.
	 */
	public static final void print(PrintStream out, List operatorList) {
		for (int index = 0; index < operatorList.size(); index++) {
			print(out, index);
			Operator operator = (Operator) operatorList.get(index);
			out.print(":");
			out.print(
				operator.isEraser()
					? "Eraser    "
					: operator.isDelimiter()
					? "Delimit"
					: operator.isAbstractor()
					? "Abstractor"
					: operator.isApplicator()
					? "Applicator"
					: operator.isDuplicator()
					? "Duplica"
					: "?");
			if (operator.isDelimiter() || operator.isDuplicator()) {
				out.print('[');
				out.print(((IndexedOperator) operator).getIndex());
				out.print(']');
			}
			out.print(" (");
			for (int number = 0; number < operator.getPortCount(); number++) {
				Port connectedPort = operator.getPort(number).getConnectedPort();
				if (connectedPort != null) {
					if (number != 0)
						out.print(" ");
					print(out, operatorList.indexOf(connectedPort.getOperator()));
					out.print('.');
					out.print(connectedPort.getNumber());
				}
			}
			out.println(")");
		}
	}

	private static final void print(PrintStream out, int n) {
		if (n < 10)
			out.print(' ');
		out.print(n);
	}
}
