/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.net.util;

import java.io.PrintStream;

import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.Port;

/**
 * Context for walking in a lambda-net exactly like traversing the read-back
 *   tree. The lambda-net is supposed to be invariant (no interaction) during
 *   walking.
 */
public interface WalkingContext {

	/**
	 * Returns the interaction net rootEraser.
	 * @return the interaction net rootEraser.
	 */
	public Eraser getRootEraser();
	
	/**
	 * Returns the currently leaved Port.
	 * @return the currently leaved Port.
	 */	
	public Port getCurrentPort();

	/**
	 * Returns the count of EndOfScope on a current bind Port.
	 * @return the count of EndOfScope on a current bind Port.
	 * @throws IllegalStateException if the current Port is not a bind Port.
	 */
	public int getEndOfScopeCount();
	
	/**
	 * Returns <tt>true</tt> if going forward, i.e. to bind.
	 * @return <tt>true</tt> if going forward, i.e. to bind.
	 */
	public boolean isWalkingForward();
	
	/**
	 * Depending on the Port connected to the currently leaved Port,
	 *   executes the following transition :
	 * <ul>
	 * <li>Abstractor.Input -> Body (forward)</li>
	 * <li>Abstractor.Body -> Input (backward)</li>
	 * <li>Abstractor.Bind -> Bind (forward -> backward)</li>
	 * <li>Applicator.Input -> Body (forward)</li>
	 * <li>Applicator.Body -> Argument (backward -> forward)</li>
	 * <li>Applicator.Argument -> Input (backward)</li>
	 * <li>Delimitor.Principal -> Auxiliary (forward or backward)</li>
	 * <li>Delimitor.Auxiliary -> Principal (forward or backward)</li>
	 * <li>Duplicator.Input -> Output (forward or backward)</li>
	 * <li>Duplicator.Output -> Input directored by context (forward or backward)</li>
	 * <li>Eraser.Port -> Port (backward -> forward)</li>
	 * </ul>
	 * @throws org.xmloperator.lambda.tree.exception.IncorrectTermException
	 *   in case of <tt>WalkingContextTranslater</tt>, if the given root Term
	 *   doesn't map to the lambda-net.
	 */	
	public void executeStep();
	
	/**
	 * Returns <tt>(currentPort.getOperator() == rootEraser)</tt>.
	 * @return <tt>(currentPort.getOperator() == rootEraser)</tt>.
	 */
	public boolean isReturnedToRoot();
	
	/**
	 * Returns to the root Eraser.
	 * <p><b>Post-condition:</b> <tt>(currentPort == rootEraser.port)</tt>.</p>
	 */
	public void returnToRoot();

	/**
	 * Executes a complete cycle and print each step in a readable fation.
	 * @param out a PrintStream.
	 */
	public void printTour(PrintStream out);
}
