/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.translate;

import java.io.PrintStream;

import org.xmloperator.lambda.net.model.Eraser;
import org.xmloperator.lambda.net.model.OperatorFactory;
import org.xmloperator.lambda.net.reduction.NetBetaUtils;
import org.xmloperator.lambda.net.reduction.NotInNormalFormException;
import org.xmloperator.lambda.net.reduction.ReductionCounter;
import org.xmloperator.lambda.net.reduction.ReductionListener;
import org.xmloperator.lambda.net.util.InteractionListener;
import org.xmloperator.lambda.net.util.InteractionStat;
import org.xmloperator.lambda.net.util.NetUtils;
import org.xmloperator.lambda.tree.generate.ChurchNumerals;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.reduction.TreeBetaUtils;
import org.xmloperator.lambda.tree.serialize.SerializeUtils;
import org.xmloperator.lambda.tree.serialize.Serializer;
import org.xmloperator.lambda.tree.util.AlphaUtils;
import org.xmloperator.lambda.tree.util.CommonTerms;
import org.xmloperator.lambda.tree.util.FactoryUtils;

/**
 * Reduces a term using a translation to a lambda-net.
 */
public abstract class Reducer {
	private static final boolean ASSERTION = Test.ASSERTION;
	private static final boolean TRACE = false;
	private static final boolean DEBUG = false;

	/**
	 * Reduces a term until the normal form, if possible.
	 * @param rootTerm a Term to reduce. It is not consumed by this method.
	 * @param operatorFactory an OperatorFactory.
	 * @param interactionListener an InteractionListener. May be null.
	 * @param redexListener a RedexListener. May be null.
	 * @return the reduced Term.
	 * @throws NotInNormalFormException if execution has been interrupted.
	 */
	public static final Term reduceDeeply(
		final Term rootTerm,
		final OperatorFactory operatorFactory,
		final InteractionListener interactionListener,
		final ReductionListener redexListener) {
		TermFactory termFactory = rootTerm.getFactory();
		Eraser rootEraser = operatorFactory.createEraser(0, false, false);
		FromTermToNet.translate(rootTerm, rootEraser);
		try {
			NetBetaUtils.reduceDeeply(
				new WalkingContextTranslaterImpl(rootEraser),
				interactionListener,
				redexListener);
			if (DEBUG)
				NetUtils.print(System.out, rootEraser);
			Term term =
				FromNetToTerm.translate(rootEraser, termFactory, interactionListener);
			if (ASSERTION && TreeBetaUtils.leftermostOutermostRedex(term) != null)
				throw new IllegalStateException();
			return term;
		}
		catch (NotInNormalFormException exception) {
			FromNetToTerm.returnNetToFactory(rootEraser);
			throw new NotInNormalFormException();
		}
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final OperatorFactory OPERATOR_FACTORY = Test.OPERATOR_FACTORY;
	private static final Serializer SERIALIZER = Test.SERIALIZER;
	private static final InteractionStat INTERACTION_STAT = Test.INTERACTION_STAT;
	private static final boolean IS_VERBOSE = false;
	private static final int N_VALUE_WITHIN_TERM_CN_C2_C2_I_I = 3;

	/**
	 * Reduce some redexes and display the results.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= TRACE | IS_VERBOSE;
			TermFactory termFactory = TERM_FACTORY;
			boolean isOk = true;
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						CommonTerms.newIdentity(termFactory),
						CommonTerms.newIdentity(termFactory)),
					CommonTerms.newIdentity(termFactory),
					"II -> I");
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						ChurchNumerals.newInt(termFactory, 2),
						ChurchNumerals.newInt(termFactory, 2)),
					ChurchNumerals.newInt(termFactory, 4),
					"(2 2) -> 4");
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						ChurchNumerals.newSucc(termFactory),
						ChurchNumerals.newInt(termFactory, 2)),
					ChurchNumerals.newInt(termFactory, 3),
					"(Succ 2) -> 3");
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							ChurchNumerals.newAdd(termFactory),
							ChurchNumerals.newInt(termFactory, 3)),
						ChurchNumerals.newInt(termFactory, 2)),
					ChurchNumerals.newInt(termFactory, 5),
					"(Add 3 2) -> 5");
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							ChurchNumerals.newMult(termFactory),
							ChurchNumerals.newInt(termFactory, 3)),
						ChurchNumerals.newInt(termFactory, 2)),
					ChurchNumerals.newInt(termFactory, 6),
					"(Mult 3 2) -> 6");
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							ChurchNumerals.newExp(termFactory),
							ChurchNumerals.newInt(termFactory, 3)),
						ChurchNumerals.newInt(termFactory, 2)),
					ChurchNumerals.newInt(termFactory, 9),
					"(Exp 3 2) -> 9");
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							ChurchNumerals.newExp(termFactory),
							ChurchNumerals.newInt(termFactory, 2)),
						ChurchNumerals.newInt(termFactory, 6)),
					ChurchNumerals.newInt(termFactory, 64),
					"(Exp 2 6) -> 64");
			{
				int n = N_VALUE_WITHIN_TERM_CN_C2_C2_I_I;
				Term term =
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							FactoryUtils.newApplication(
								FactoryUtils.newApplication(
									ChurchNumerals.newInt(termFactory, n),
									ChurchNumerals.newInt(termFactory, 2)),
								ChurchNumerals.newInt(termFactory, 2)),
							CommonTerms.newIdentity(termFactory)),
						CommonTerms.newIdentity(termFactory));
				InteractionStat interactionStat = new InteractionStat();
				long t0 = System.currentTimeMillis();
				Term result =
					reduceDeeply(term, OPERATOR_FACTORY, interactionStat, null);
				long t1 = System.currentTimeMillis();
				FactoryUtils.returnTerm(term);
				FactoryUtils.returnTerm(result);
				if (isVerbose)
					System.out.println(
						"Reducer ("
							+ n
							+ " 2 2 I I) : "
							+ interactionStat
							+ " using "
							+ (t1 - t0)
							+ " ms");
			}
			{
				Term term = CommonTerms.newOmega(termFactory);
				ReductionListener betaCounter = new ReductionCounter(10);
				try {
					Reducer.reduceDeeply(
						term,
						OPERATOR_FACTORY,
						INTERACTION_STAT,
						betaCounter);
				}
				catch (NotInNormalFormException exception) {
				}
				FactoryUtils.returnTerm(term);
			}
			return isOk;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static final boolean test(
		PrintStream out,
		boolean isVerbose,
		Term term,
		Term expectedResult,
		String operation) {
		ReductionCounter reductionCounter = new ReductionCounter(100);
		if (TRACE)
			reductionCounter.nextReductionListener =
				new ReductionTracer(SERIALIZER, true, true, term.getFactory());
		Term result =
			reduceDeeply(term, OPERATOR_FACTORY, INTERACTION_STAT, reductionCounter);
		if (TRACE)
			SerializeUtils.printTerm(SERIALIZER, result);
		boolean isOk = AlphaUtils.strictlyEquals(result, expectedResult);
		if (!isOk) {
			System.out.print("result  :");
			SerializeUtils.printTerm(SERIALIZER, result);
			System.out.print("expected:");
			SerializeUtils.printTerm(SERIALIZER, expectedResult);
		}
		FactoryUtils.returnTerm(term);
		FactoryUtils.returnTerm(result);
		FactoryUtils.returnTerm(expectedResult);
		if (isVerbose || !isOk) {
			out.print("Reducer ");
			out.print(isOk ? "OK" : "KO");
			out.print(" ");
			out.print(operation);
			out.print(" using ");
			out.print(reductionCounter.getCount());
			out.print(" step(s).");
			out.println();
		}
		return isOk;
	}
}
