/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.generate;

import java.io.PrintStream;

import org.xmloperator.lambda.tree.exception.NotInNormalFormException;
import org.xmloperator.lambda.tree.exception.TermException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.reduction.TreeBetaUtils;
import org.xmloperator.lambda.tree.reduction.ReductionCounter;
import org.xmloperator.lambda.tree.reduction.ReductionTracer;
import org.xmloperator.lambda.tree.serialize.SerializeUtils;
import org.xmloperator.lambda.tree.serialize.Serializer;
import org.xmloperator.lambda.tree.util.CommonTerms;
import org.xmloperator.lambda.tree.util.FactoryUtils;
import org.xmloperator.lambda.tree.util.SignatureUtils;
import org.xmloperator.lambda.tree.util.WellFormedness;

/**
 * Factories for Church numerals.
 */
public abstract class ChurchNumerals {
	private static final boolean ASSERTION = Test.ASSERTION;
	private static final boolean TRACE = false;
	
	/**
	 * Creates and returns a new instance of the Abstraction that represents
	 *   the integer n:
	 * <ul>
	 * <li>0 = "aav",</li>
	 * <li>1 = "aacevv",</li>
	 * <li>2 = "aacevcevv",</li>
	 * <li>3 = "aacevcevcevv",</li>
	 * <li>...</li>
	 * </ul>
	 * @param termFactory a TermFactory.
	 * @param n the integer to represent.
	 * @return the Church numeral Abstraction.
	 * @throws IllegalArgumentException if the integer n is negative.
	 */
	public static final Abstraction newInt(TermFactory termFactory, int n) {
		if (n < 0)
			throw new IllegalArgumentException();
		Term term = termFactory.createLeaf();
		for (int k = 0; k < n; k++)
			term =
				FactoryUtils.newApplication(FactoryUtils.newLeaf(termFactory, 1), term);
		return FactoryUtils.newAbstraction(FactoryUtils.newAbstraction(term));
	}

	/**
	 * Returns the integer value of a Church numeral or -1 if the argument is not
	 *   a Church numeral.
	 * @param abstraction the Church numeral.
	 * @return the value of the Church numeral or -1 if the argument is not
	 *   a Church numeral.
	 * @throws NullPointerException if the given Abstraction is broken.
	 */
	public static final int intValue(final Abstraction abstraction) {
		Term body = abstraction.getBody();
		if (!body.isAbstraction())
			return -1;
		Term argument = ((Abstraction) body).getBody();
		int value = 0;
		while (true) {
			if (argument.isLeaf())
				return value;
			if (!argument.isApplication())
				return -1;
			Application application = (Application) argument;
			body = application.getBody();
			if (!body.isEndOfScope())
				return -1;
			if (!((EndOfScope) body).getBody().isLeaf())
				return -1;
			argument = application.getArgument();
			value++;
		}
	}

	/**
	 * Creates and returns a new instance of the "Succ" Abstraction
	 *   ("aaacevcceevevv").
	 * @param termFactory an TermFactory.
	 * @return the "Succ" Abstraction.
	 */
	public static final Abstraction newSucc(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newAbstraction(
					FactoryUtils.newApplication(
						FactoryUtils.newLeaf(termFactory, 1),
						FactoryUtils.newApplication(
							FactoryUtils.newApplication(
								FactoryUtils.newLeaf(termFactory, 2),
								FactoryUtils.newLeaf(termFactory, 1)),
							termFactory.createLeaf())))));
	}

	/**
	 * Creates and returns a new instance of the "Zero" Abstraction
	 *   ("aaacceevcKvev").
	 * @param termFactory an TermFactory.
	 * @return the "Zero" Abstraction.
	 */
	public static final Abstraction newZero(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newAbstraction(
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							FactoryUtils.newLeaf(termFactory, 2),
							FactoryUtils.newApplication(
								CommonTerms.newTrue(termFactory),
								termFactory.createLeaf())),
						FactoryUtils.newLeaf(termFactory, 1)))));
	}

	/**
	 * Creates and returns a new instance of the "Add" Abstraction
	 *   ("aaaacceeevevcceevevv").
	 * @param termFactory an TermFactory.
	 * @return the "Add" Abstraction.
	 */
	public static final Abstraction newAdd(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newAbstraction(
					FactoryUtils.newAbstraction(
						FactoryUtils.newApplication(
							FactoryUtils.newApplication(
								FactoryUtils.newLeaf(termFactory, 3),
								FactoryUtils.newLeaf(termFactory, 1)),
							FactoryUtils.newApplication(
								FactoryUtils.newApplication(
									FactoryUtils.newLeaf(termFactory, 2),
									FactoryUtils.newLeaf(termFactory, 1)),
								termFactory.createLeaf()))))));
	}

	/**
	 * Creates and returns a new instance of the "Mult" Abstraction
	 *   ("aaaceevcevv").
	 * @param termFactory an TermFactory.
	 * @return the "Mult" Abstraction.
	 */
	public static final Abstraction newMult(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newAbstraction(
					FactoryUtils.newApplication(
						FactoryUtils.newLeaf(termFactory, 2),
						FactoryUtils.newApplication(
							FactoryUtils.newLeaf(termFactory, 1),
							termFactory.createLeaf())))));
	}

	/**
	 * Creates and returns a new instance of the "Exp" Abstraction ("aacvev").
	 * @param termFactory an TermFactory.
	 * @return the "Exp" Abstraction.
	 */
	public static final Abstraction newExp(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newApplication(
					termFactory.createLeaf(),
					FactoryUtils.newLeaf(termFactory, 1))));
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final Serializer SERIALIZER = Test.SERIALIZER;
	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final boolean IS_VERBOSE = false;
	private static final int N_VALUE_WITHIN_TERM_CN_C2_C2_I_I = 2;

	/**
	 * Execute some operations using Church numerals.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= TRACE | IS_VERBOSE;
			TermFactory termFactory = TERM_FACTORY;
			boolean isOk = true;
			for (int k = 0; k < 10; k++) {
				Term term = newInt(termFactory, k);
				WellFormedness.verifyWellFormedness(term);
				FactoryUtils.returnTerm(term);
			}
			isOk &= test(out, isVerbose, "(0)", newInt(termFactory, 0), 0, 10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Succ 0)",
					FactoryUtils.newApplication(
						newSucc(termFactory),
						newInt(termFactory, 0)),
					1,
					10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Succ 1)",
					FactoryUtils.newApplication(
						newSucc(termFactory),
						newInt(termFactory, 1)),
					2,
					10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Succ 2)",
					FactoryUtils.newApplication(
						newSucc(termFactory),
						newInt(termFactory, 2)),
					3,
					10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Add 3 2)",
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							newAdd(termFactory),
							newInt(termFactory, 3)),
						newInt(termFactory, 2)),
					5,
					10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Mult 3 2)",
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							newMult(termFactory),
							newInt(termFactory, 3)),
						newInt(termFactory, 2)),
					6,
					20);
			isOk
				&= test(
					out,
					isVerbose,
					"(2 2)",
					FactoryUtils.newApplication(
						newInt(termFactory, 2),
						newInt(termFactory, 2)),
					4,
					10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Exp 3 2)",
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							newExp(termFactory),
							newInt(termFactory, 3)),
						newInt(termFactory, 2)),
					9,
					10);
			isOk
				&= test(
					out,
					isVerbose,
					"(Exp 2 4)",
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							newExp(termFactory),
							newInt(termFactory, 2)),
						newInt(termFactory, 4)),
					16,
					1000);
			{
				int n = N_VALUE_WITHIN_TERM_CN_C2_C2_I_I;
				Term term =
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							FactoryUtils.newApplication(
								FactoryUtils.newApplication(
									newInt(termFactory, n),
									newInt(termFactory, 2)),
								newInt(termFactory, 2)),
							CommonTerms.newIdentity(termFactory)),
						CommonTerms.newIdentity(termFactory));
				try {
					long t0 = System.currentTimeMillis();
					ReductionCounter reductionCounter = new ReductionCounter(1000000);
					if (TRACE)
						reductionCounter.nextReductionListener =
							new ReductionTracer(SERIALIZER, true, true, term);
					term = TreeBetaUtils.reduceDeeply(term, reductionCounter);
					long t1 = System.currentTimeMillis();
					if (isVerbose)
						out.println(
							"ChurchNumerals ("
								+ n
								+ " 2 2 I I) : "
								+ reductionCounter.getCount()
								+ " steps in "
								+ (t1 - t0)
								+ " ms");
				}
				catch (NotInNormalFormException exception) {
					term = exception.getTerm();
				}
				FactoryUtils.returnTerm(term);
			}
			return isOk;
		}
		catch (TermException exception) {
			System.out.println("ChurchNumerals TermException");
			Term subTerm = exception.getTerm();
			if (SERIALIZER != null && subTerm != null)
				SerializeUtils.printUnderlinedTerm(SERIALIZER, null, subTerm);
			exception.printStackTrace(out);
			return false;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static final boolean test(
		PrintStream out,
		boolean isVerbose,
		String termName,
		Term term,
		int expectedResultValue,
		int maxCount) {
		int termSize = SignatureUtils.leafAndAbstractionCount(term);
		ReductionCounter reductionCounter = new ReductionCounter(maxCount);
		if (TRACE)
			reductionCounter.nextReductionListener =
				new ReductionTracer(SERIALIZER, true, true, term);
		Term result = TreeBetaUtils.reduceDeeply(term, reductionCounter);
		int resultIntValue = -1;
		WellFormedness.verifyWellFormedness(result);
		resultIntValue = intValue((Abstraction) result);
		FactoryUtils.returnTerm(result);
		boolean isOk = resultIntValue == expectedResultValue;
		if (isVerbose || !isOk) {
			out.print("ChurchNumerals ");
			out.print(isOk ? "OK" : "KO");
			out.print(" ");
			out.print(termName);
			out.print("[" + termSize + "]");
			out.print(" == ");
			out.print(resultIntValue + "/" + expectedResultValue);
			out.print(" using " + reductionCounter.getCount() + " steps.");
			out.println();
		}
		return isOk;
	}
}
