/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.generate;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.xmloperator.lambda.tree.exception.NotInNormalFormException;
import org.xmloperator.lambda.tree.exception.TermException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.Leaf;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.reduction.TreeBetaUtils;
import org.xmloperator.lambda.tree.reduction.ReductionCounter;
import org.xmloperator.lambda.tree.reduction.ReductionTracer;
import org.xmloperator.lambda.tree.serialize.SerializeUtils;
import org.xmloperator.lambda.tree.serialize.Serializer;
import org.xmloperator.lambda.tree.util.CommonTerms;
import org.xmloperator.lambda.tree.util.FactoryUtils;
import org.xmloperator.lambda.tree.util.SubstitutionUtils;

/**
 * Generator of random closed Terms.
 */
public class RandomGenerator implements TermIterator {
	private static final boolean TRACE = false;

	private final TermFactory termFactory;
	private final Random random;
	private final int treeSize;

	public RandomGenerator(TermFactory termFactory, long seed, int treeSize) {
		if (termFactory == null || treeSize < 2)
			throw new IllegalArgumentException();
		this.termFactory = termFactory;
		this.random = new Random(seed);
		this.treeSize = treeSize;
	}

	private Term createTerm(int treeSize) {
		if (treeSize < 2)
			throw new IllegalArgumentException();
		if (treeSize < 4 || this.random.nextBoolean())
			return this.createAbstraction(treeSize);
		int bodyStepCount = this.random.nextInt(treeSize - 3);
		Term term =
			FactoryUtils.newApplication(
				this.createTerm(2 + bodyStepCount),
				this.createTerm(treeSize - 2 - bodyStepCount));
		return term;
	}

	/**
	 * @return a random closed Abstraction, having a given tree size.
	 * @throws an IllegalArgumentException if treeSize < 2.
	 */
	public Abstraction createAbstraction(int treeSize) {
		if (treeSize < 2)
			throw new IllegalArgumentException();
		Abstraction abstraction = CommonTerms.newIdentity(termFactory);
		List leafList = new ArrayList();
		leafList.add(abstraction.getBody());
		for (int k = 0; k < treeSize - 2; k++)
			this.executeStep(leafList);
		for (int k = 0; k < leafList.size(); k++) {
			Leaf leaf = (Leaf) leafList.get(k);
			int abstractionCount = abstractionCount(leaf);
			insertEndOfScopes(leaf, this.random.nextInt(abstractionCount));
		}
		return abstraction;
	}

	private void executeStep(List leafList) {
		int leafIndex = this.random.nextInt(leafList.size());
		Leaf leaf = (Leaf) leafList.get(leafIndex);
		if (this.random.nextBoolean())
			SubstitutionUtils.insert(this.termFactory.createAbstraction(), leaf);
		else {
			Application application = this.termFactory.createApplication();
			SubstitutionUtils.insert(application, leaf);
			Leaf newLeaf = this.termFactory.createLeaf();
			application.setArgument(newLeaf);
			leafList.add(newLeaf);
		}
	}

	private static final int abstractionCount(final Leaf leaf) {
		int count = 0;
		Term parent = leaf.getParent();
		while (parent != null) {
			if (parent.isAbstraction())
				count++;
			parent = parent.getParent();
		}
		return count;
	}

	private static final void insertEndOfScopes(final Leaf leaf, int count) {
		TermFactory termFactory = leaf.getFactory();
		for (int k = 0; k < count; k++)
			SubstitutionUtils.insert(termFactory.createEndOfScope(), leaf);
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final Serializer SERIALIZER = Test.SERIALIZER;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Executes some Term generations.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= TRACE | IS_VERBOSE;
			TermFactory termFactory = TERM_FACTORY;
			long seed = 12345;
			int treeSize = 9;
			RandomGenerator generator =
				new RandomGenerator(termFactory, seed, treeSize);
			int count = 100;
			boolean isOk = true;
			for (int k = 0; k <= count; k++)
				isOk &= test(out, isVerbose, generator.next());
			return isOk;
		}
		catch (TermException exception) {
			System.out.println("RandomTermGenerator TermException");
			Term subTerm = exception.getTerm();
			if (SERIALIZER != null && subTerm != null)
				SerializeUtils.printUnderlinedTerm(SERIALIZER, null, subTerm);
			exception.printStackTrace(out);
			return false;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static final boolean test(
		PrintStream out,
		boolean isVerbose,
		Term term) {
		if (true || TreeBetaUtils.leftermostOutermostRedex(term) != null) {
			if (TRACE) {
				System.out.print("---");
				SerializeUtils.printTerm(SERIALIZER, term);
			}
			ReductionCounter reductionCounter = new ReductionCounter(5);
			if (TRACE)
				reductionCounter.nextReductionListener =
					new ReductionTracer(SERIALIZER, false, false, term);
			try {
				term = TreeBetaUtils.reduceDeeply(term, reductionCounter);
			}
			catch (NotInNormalFormException exception) {
				term = exception.getTerm();
			}
		}
		FactoryUtils.returnTerm(term);
		return true;
	}

	////////////////////////////////////
	// TermIterator implementation
	////////////////////////////////////

	public boolean hasNext() {
		return true;
	}

	public Term next() {
		return this.createTerm(this.treeSize);
	}
}
