/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.impl;

import org.xmloperator.lambda.shared.ObjectArrayFactory;
import org.xmloperator.lambda.shared.Stack;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Leaf;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;

/**
 * TermFactory implementation.
 * Can either reuse objects or garbage them.
 */
public final class TermFactoryImpl implements TermFactory {
	private final boolean areObjectsReused;
	private final Stack leafStack = new Stack();
	private final Stack endOfScopeStack = new Stack();
	private final Stack abstractionStack = new Stack();
	private final Stack applicationStack = new Stack();
	private final ObjectArrayFactory objectArrayFactory;
	private final Stack termStackStack = new Stack();
	private int newLeafCount = 0;
	private int newEndOfScopeCount = 0;
	private int newAbstractionCount = 0;
	private int newApplicationCount = 0;
	private int newTermStackCount = 0;

	/**
	 * TermFactory constructor.
	 * @param areObjectsReused:
	 * 	 - true: returned objects are reused.
	 *   - false: returned objects are garbaged.
	 * @param objectArrayFactory an ObjectArrayFactory.
	 */
	public TermFactoryImpl(
		boolean areObjectsReused,
		ObjectArrayFactory objectArrayFactory) {
		this.areObjectsReused = areObjectsReused;
		this.objectArrayFactory = objectArrayFactory;
	}

	/**
	 * Clear memory.
	 */
	public void clear() {
		this.leafStack.clear();
		this.endOfScopeStack.clear();
		this.abstractionStack.clear();
		this.applicationStack.clear();
		this.termStackStack.clear();
		this.newLeafCount = 0;
		this.newAbstractionCount = 0;
		this.newEndOfScopeCount = 0;
		this.newApplicationCount = 0;
		this.newTermStackCount = 0;
	}

	/**
	 * @return true if all the objects that have been created have been returned.
	 */
	public boolean isWellBalanced() {
		return this.leafStack.getSize() == this.newLeafCount
			&& this.endOfScopeStack.getSize() == this.newEndOfScopeCount
			&& this.abstractionStack.getSize() == this.newAbstractionCount
			&& this.applicationStack.getSize() == this.newApplicationCount
			&& this.termStackStack.getSize() == this.newTermStackCount;
	}

	public String toString() {
		return "[Leaf("
			+ this.newLeafCount
			+ ","
			+ this.leafStack.getSize()
			+ "), EndOfScope("
			+ this.newEndOfScopeCount
			+ ","
			+ this.endOfScopeStack.getSize()
			+ "), Abstraction("
			+ this.newAbstractionCount
			+ ","
			+ this.abstractionStack.getSize()
			+ "), Application("
			+ this.newApplicationCount
			+ ","
			+ this.applicationStack.getSize()
			+ "), TermStack("
			+ this.newTermStackCount
			+ ","
			+ this.termStackStack.getSize()
			+ ")]";
	}

	void returnLeaf(Leaf leaf) {
		if (this.areObjectsReused)
			this.leafStack.push(leaf);
	}

	void returnEndOfScope(EndOfScope endOfScope) {
		if (endOfScope.getFactory() != this)
			throw new IllegalArgumentException();
		if (this.areObjectsReused)
			this.endOfScopeStack.push(endOfScope);
	}

	void returnAbstraction(Abstraction abstraction) {
		if (abstraction.getFactory() != this)
			throw new IllegalArgumentException();
		if (this.areObjectsReused)
			this.abstractionStack.push(abstraction);
	}

	void returnApplication(Application application) {
		if (application.getFactory() != this)
			throw new IllegalArgumentException();
		if (this.areObjectsReused)
			this.applicationStack.push(application);
	}

	void returnTermStack(TermStack termStack) {
		if (this.areObjectsReused)
			this.termStackStack.push(termStack);
	}

	/////////////////////////////
	// TermFactory implementation
	/////////////////////////////

	public Leaf createLeaf() {
		if (this.areObjectsReused) {
			Leaf leaf = (Leaf) this.leafStack.pop();
			if (leaf != null) {
				((LeafImpl) leaf).isReturnedToTheFactory = false;
				return leaf;
			}
		}
		this.newLeafCount++;
		return new LeafImpl(this);
	}

	public EndOfScope createEndOfScope() {
		if (this.areObjectsReused) {
			EndOfScope endOfScope = (EndOfScope) this.endOfScopeStack.pop();
			if (endOfScope != null) {
				((EndOfScopeImpl) endOfScope).isReturnedToTheFactory = false;
				return endOfScope;
			}
		}
		this.newEndOfScopeCount++;
		return new EndOfScopeImpl(this);
	}

	public Abstraction createAbstraction() {
		if (this.areObjectsReused) {
			Abstraction abstraction = (Abstraction) this.abstractionStack.pop();
			if (abstraction != null) {
				((AbstractionImpl) abstraction).isReturnedToTheFactory = false;
				return abstraction;
			}
		}
		this.newAbstractionCount++;
		return new AbstractionImpl(this);
	}

	public Application createApplication() {
		if (this.areObjectsReused) {
			Application application = (Application) this.applicationStack.pop();
			if (application != null) {
				((ApplicationImpl) application).isReturnedToTheFactory = false;
				return application;
			}
		}
		this.newApplicationCount++;
		return new ApplicationImpl(this);
	}

	public TermStack createTermStack(boolean acceptNull) {
		if (this.areObjectsReused) {
			TermStackImpl termStack = (TermStackImpl) this.termStackStack.pop();
			if (termStack != null) {
				termStack.acceptNull = acceptNull;
				termStack.startUse();
				return termStack;
			}
		}
		this.newTermStackCount++;
		TermStackImpl termStack =
			new TermStackImpl(this, this.objectArrayFactory, acceptNull);
		termStack.startUse();
		return termStack;
	}
}
