/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.model;

/**
 * Namefree lambda-term.
 * This is either a <tt>Leaf</tt>, an <tt>EndOfScope</tt>,
 *   an <tt>Abstraction</tt> or an <tt>Application</tt>. A Term instance is
 *   created by using a <tt>TermFactory</tt> and is terminated by being returned
 *   to the same <tt>TermFactory</tt>.
 * @see org.xmloperator.lambda.tree.model.TermFactory
 */
public interface Term {

	/**
	 * Returns the factory of this Term, which can be used for building
	 *   compatible Terms.
	 * @return the factory of this Term.
	 */
	public TermFactory getFactory();

	/**
	 * Returns <tt>true</tt> if this can be returned to its factory.
	 * If this has a parent or a content then this cannot be returned to factory.
	 * @return <tt>true</tt> if this can be returned to its factory.
	 */
	public boolean canBeReturnedToFactory();
	
	/**
	 * Returns this Term to its Factory.
	 * <p><b>Pre-condition:</b> can be returned to factory.</p>
	 * <p><b>Post-condition:</b> is no more usable.</p>
	 * @throws IllegalStateException if this cannot be returned to factory.
	 */
	public void returnToFactory();
	
	public static final int TYPE_LEAF = 0;
	public static final int TYPE_ENDOFSCOPE = 1;
	public static final int TYPE_ABSTRACTION = 2;
	public static final int TYPE_APPLICATION = 3;

	/**
	 * Returns the type of this term.
	 * @return the type of this term. It is one of the constants TYPE_*
	 */	
	public int getType();
	
	/**
	 * Returns <tt>true</tt> if this Term is a Leaf.
	 * @return <tt>true</tt> if this Term is a Leaf.
	 */
	public boolean isLeaf();
	
	/**
	 * Returns <tt>true</tt> if this Term is an EndOfScope.
	 * @return <tt>true</tt> if this Term is an EndOfScope.
	 */
	public boolean isEndOfScope();
	
	/**
	 * Returns <tt>true</tt> if this Term is an Abstraction.
	 * @return <tt>true</tt> if this Term is an Abstraction.
	 */
	public boolean isAbstraction();
	
	/**
	 * Returns <tt>true</tt> if this Term is an Application.
	 * @return <tt>true</tt> if this Term is an Application.
	 */
	public boolean isApplication();

	/**
	 * A Term may have a parent Term.
	 * A parent Term is either an <tt>EndOfScope</tt>, an <tt>Abstraction</tt>,
	 *   or an <tt>Application</tt>.
	 * Returns the parent of this Term, if any.
	 * @return the parent of this Term if any, null elsewhere.
	 */
	public TermWithBody getParent();
}
