/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.reduction;

import java.io.PrintStream;

import org.xmloperator.lambda.tree.exception.NotInNormalFormException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.model.TermWithBody;
import org.xmloperator.lambda.tree.serialize.Serializer;
import org.xmloperator.lambda.tree.util.AlphaUtils;
import org.xmloperator.lambda.tree.util.CommonTerms;
import org.xmloperator.lambda.tree.util.FactoryUtils;
import org.xmloperator.lambda.tree.util.SignatureUtils;
import org.xmloperator.lambda.tree.util.SubstitutionUtils;
import org.xmloperator.lambda.tree.util.WellFormedness;

/**
 * Utilities about beta reduction.
 */
public abstract class TreeBetaUtils {
	private static final boolean ASSERTION = Test.ASSERTION;
	private static final boolean TRACE = false;

	/**
	 * Returns the leftermost-outermost redex of a Term or null if the Term has
	 *   no redex.
	 * <p><b>Pre-condition:</b> the given Term is normalized about end-of-scopes.
	 *   </p>
	 * <p>Normalisation about EndOfScopes, i.e. scope extrusion, can be verified
	 *   by using {@link AlphaUtils#verifyScopeExtrusion(Term)}.</p>
	 * @param rootTerm a Term.
	 * @return the leftermost-outermost redex of the given Term or null
	 *   if it has no redex.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final Application leftermostOutermostRedex(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf());
				else if (term.isAbstraction() || term.isEndOfScope()) {
					term = ((TermWithBody) term).getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					term = application.getBody();
					if (term.isAbstraction())
						return application;
					termStack.push(application.getArgument());
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					return null;
				term = termStack.pop();
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	/**
	 * Reduces a redex.
	 * Returns the resulting Term.
	 * Returns the redex Application and the redex Abstraction to the factory.
	 * <p><b>Pre-condition:</b> the redex is normalized about EndOfScopes.</p>
	 * <p><b>Post-condition:</b> the redex is normalized about EndOfScopes.</p>
	 * <p>Normalisation about EndOfScopes, i.e scope extrusion, can be verified
	 *   by using {@link AlphaUtils#verifyScopeExtrusion(Term)}.</p>
	 * @param redex a redex to reduce.
	 * @return the resulting Term. This is the body of the redex Abstraction.
	 * @throws IllegalArgumentException if the given redex is not a redex.
	 * @throws NullPointerException if the given redex is broken.
	 */
	public static final Term reduce(final Application redex) {
		if (ASSERTION)
			AlphaUtils.verifyScopeExtrusion(redex);
		Term body = redex.getBody();
		if (!body.isAbstraction())
			throw new IllegalArgumentException();
		Abstraction redexAbstraction = (Abstraction) body;
		Term argument = redex.getArgument();
		redex.clearBody();
		redex.clearArgument();
		body = redexAbstraction.getBody();
		redexAbstraction.clearBody();
		SubstitutionUtils.substitute(redex, body);
		TermFactory termFactory = redex.getFactory();
		redex.returnToFactory();
		redexAbstraction.returnToFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			int contextSize = 0;
			// contextSize is the count of visible Abstraction(s),
			//   redex Abstraction excepted.
			Term term = body;
			while (true) {
				if (term.isLeaf()) {
					if (contextSize == 0 && !argument.isLeaf()) {
						Term cloneArgument = AlphaUtils.clone(argument);
						SubstitutionUtils.substitute(term, cloneArgument);
						if (body == term)
							body = cloneArgument;
						term.returnToFactory();
						if (!argument.isEndOfScope())
							while (true) {
								Term parent = cloneArgument.getParent();
								if (parent == null || !parent.isEndOfScope())
									break;
								AlphaUtils.extrude((EndOfScope) parent);
							}
					}
				}
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					if (contextSize == 0) {
						SubstitutionUtils.removeTermWithoutItsBody(endOfScope);
						endOfScope.returnToFactory();
					}
					else {
						contextSize--;
						termStack.push(endOfScope);
						term = endOfScope.getBody();
						continue;
					}
				}
				else if (term.isAbstraction()) {
					Abstraction abstraction = (Abstraction) term;
					contextSize++;
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty()) {
						term = null;
						break;
					}
					term = termStack.pop();
					if (term.isEndOfScope())
						contextSize++;
					else if (term.isAbstraction())
						contextSize--;
					else if (term.isApplication()) {
						Application application = (Application) term;
						term = application.getArgument();
						break;
					}
					else
						throw new IllegalStateException();
				}
				if (term == null)
					break;
			}
			FactoryUtils.returnTerm(argument);
		}
		finally {
			termStack.returnToFactory();
		}
		return body;
	}

	/**
	 * Reduces a Term until a normal form is obtained.
	 * The process may be aborted by a <tt>NotInNormalFormException</tt> thrown
	 *   by {@link ReductionListener#notifyStartReduction}.
	 * @param rootTerm a Term to reduce.
	 * @param reductionListener a listener about beta-reduction process.
	 *   May be null.
	 * @return the resulting Term.
	 * @throws NullPointerException if the given root term is broken.
	 * @throws NotInNormalFormException if the ReductionListener do so.
	 *   The obtained Term is provided by the exception.
	 */
	public static final Term reduceDeeply(
		Term rootTerm,
		final ReductionListener reductionListener) {
		while (true) {
			Application redex = leftermostOutermostRedex(rootTerm);
			if (redex == null)
				break;
			boolean isRoot = redex == rootTerm;
			if (reductionListener != null)
				try {
					reductionListener.notifyStartReduction(redex);
				}
				catch (NotInNormalFormException exception) {
					exception.setTerm(rootTerm);
					throw exception;
				}
			Term result = reduce(redex);
			if (reductionListener != null)
				reductionListener.notifyEndReduction(result);
			if (isRoot)
				rootTerm = result;
			if (ASSERTION)
				WellFormedness.verifyWellFormedness(rootTerm);
		}
		return rootTerm;
	}

	public static void main(String[] args) {
		if (test(System.out, false))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final Serializer SERIALIZER = Test.SERIALIZER;
	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Reduce some redexes and display the results.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= TRACE | IS_VERBOSE;
			TermFactory termFactory = TERM_FACTORY;
			boolean isOk = true;
			isOk
				&= test(
					out,
					isVerbose,
					FactoryUtils.newApplication(
						CommonTerms.newIdentity(termFactory),
						CommonTerms.newIdentity(termFactory)),
					CommonTerms.newIdentity(termFactory),
					"II",
					"I");
			isOk
				&= test(
					out,
					isVerbose,
					CommonTerms.newOmega(termFactory),
					CommonTerms.newOmega(termFactory),
					"Omega",
					"Omega");
			isOk
				&= test(
					out,
					isVerbose,
					(Application) reduce(CommonTerms.newOmega(termFactory)),
					CommonTerms.newOmega(termFactory),
					"Omega",
					"Omega");
			{
				long t0 = System.currentTimeMillis();
				Term term = CommonTerms.newOmega(termFactory);
				int maxCount = 2;
				ReductionCounter reductionCounter = new ReductionCounter(maxCount);
				if (TRACE)
					reductionCounter.nextReductionListener =
						new ReductionTracer(SERIALIZER, true, true, term);
				try {
					term = reduceDeeply(term, reductionCounter);
				}
				catch (NotInNormalFormException exception) {
					term = exception.getTerm();
				}
				FactoryUtils.returnTerm(term);
				long t1 = System.currentTimeMillis();
				if (isVerbose)
					out.println(
						"BetaUtils "
							+ maxCount
							+ " Omega reductions uses "
							+ (t1 - t0)
							+ " ms");
			}
			return isOk;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private static final boolean test(
		PrintStream out,
		boolean isVerbose,
		Application redex,
		Term expectedResult,
		String redexName,
		String resultName) {
		int redexSize = -1;
		int resultSize = -1;
		WellFormedness.verifyWellFormedness(redex);
		redexSize = SignatureUtils.leafAndAbstractionCount(redex);
		Term result = reduce(redex);
		resultSize = SignatureUtils.leafAndAbstractionCount(result);
		boolean isOk = AlphaUtils.strictlyEquals(result, expectedResult);
		WellFormedness.verifyWellFormedness(result);
		FactoryUtils.returnTerm(result);
		FactoryUtils.returnTerm(expectedResult);
		if (isVerbose || !isOk) {
			out.print("BetaUtils ");
			out.print(isOk ? "OK" : "KO");
			out.print(" ");
			out.print(redexName);
			out.print("(" + redexSize + ")");
			out.print(" -> ");
			out.print(resultName);
			out.print("(" + resultSize + ")");
			out.println();
		}
		return isOk;
	}
}
