/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.serialize;

import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Reader;

import org.xmloperator.lambda.tree.exception.ParsingException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Leaf;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.model.TermWithBody;
import org.xmloperator.lambda.tree.util.FactoryUtils;

/**
 * Prefixed Expression serializer designed for storage purpose
 *   but also convenient for a human reader.
 * <p>--- Term syntax (API style)</p>
 * <pre>
 * Term ::= Leaf | EndOfScope | Abstraction | Application
 * Leaf ::= 'v'
 * EndOfScope ::= 'e' Term
 * Abstraction ::= 'a' Term
 * Application ::= 'c' Term Term
 * </pre>
 */
public class PrefixedSerializer implements Serializer {
	public static final String STYLE_API = "veac";
	public static final String STYLE_SCOPE = "0/\\@";
	static final int STYLE_POS_LEAF = 0;
	static final int STYLE_POS_ENDOFSCOPE = 1;
	static final int STYLE_POS_ABSTRACTION = 2;
	static final int STYLE_POS_APPLICATION = 3;
	static final char SPACE_DEFAULT_CHAR = ' ';
	static final char UNDERLINE_DEFAULT_CHAR = '=';
	static final char REDEX_DEFAULT_CHAR = '-';
	static final char REDEX_ABSTRACTION_DEFAULT_CHAR = '+';
	static final char REDEX_BINDING_DEFAULT_CHAR = '^';
	static final char REDEX_ARGUMENT_DEFAULT_CHAR = '*';
	static final int UNDERLINE_OUT = 0;
	static final int UNDERLINE_IN = 1;
	static final int UNDERLINE_REDEX = 2;
	static final int UNDERLINE_REDEX_END_OF_SCOPE = 3;
	static final int UNDERLINE_REDEX_ABSTRACTION = 4;
	static final int UNDERLINE_REDEX_FUNCTION = 5;
	static final int UNDERLINE_REDEX_ARGUMENT = 6;

	private final char leafChar;
	private final char endOfScopeChar;
	private final char abstractionChar;
	private final char applicationChar;
	private final char spaceChar = SPACE_DEFAULT_CHAR;
	private final char underlineChar = UNDERLINE_DEFAULT_CHAR;
	private final char redexChar = REDEX_DEFAULT_CHAR;
	private final char redexAbstractionChar = REDEX_ABSTRACTION_DEFAULT_CHAR;
	private final char redexBindingChar = REDEX_BINDING_DEFAULT_CHAR;
	private final char redexArgumentChar = REDEX_ARGUMENT_DEFAULT_CHAR;
	private final char[] underlineChars;

	public PrefixedSerializer() {
		this(STYLE_API);
	}

	public PrefixedSerializer(String style) {
		if (style.length() != 4)
			throw new IllegalArgumentException();
		this.leafChar = style.charAt(STYLE_POS_LEAF);
		this.endOfScopeChar = style.charAt(STYLE_POS_ENDOFSCOPE);
		this.abstractionChar = style.charAt(STYLE_POS_ABSTRACTION);
		this.applicationChar = style.charAt(STYLE_POS_APPLICATION);
		this.underlineChars =
			new char[] {
				this.spaceChar,
				this.redexChar,
				this.redexChar,
				this.redexChar,
				this.redexAbstractionChar,
				this.redexChar,
				this.redexArgumentChar };
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final String VALUE_I = "av";
	private static final String VALUE_K = "aaev";
	private static final String VALUE_F = "aeav";
	private static final String VALUE_C = "aaaccveevev";
	private static final String VALUE_FST = "acve" + VALUE_K;
	private static final String VALUE_SND = "acve" + VALUE_F;
	private static final String VALUE_DELTA = "acvv";
	private static final String VALUE_OMEGA = "c" + VALUE_DELTA + VALUE_DELTA;
	private static final String VALUE_Y = "acacevcvvacevcvv";
	private static final String VALUE_V = "aacvcecvvv";
	private static final String VALUE_TETA = "c" + VALUE_V + VALUE_V;
	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Serializes in and out some terms.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			String[] values =
				{
					VALUE_I,
					VALUE_K,
					VALUE_F,
					VALUE_C,
					VALUE_FST,
					VALUE_SND,
					VALUE_SND,
					VALUE_OMEGA,
					VALUE_Y,
					VALUE_Y,
					VALUE_TETA };
			TermFactory expressionFactory = TERM_FACTORY;
			boolean isAllOk = true;
			for (int index = 0; index < values.length; index++) {
				boolean isOk = false;
				String value = values[index];
				try {
					Serializer serializer = new PrefixedSerializer();
					Term term =
						SerializeUtils.stringToTerm(serializer, expressionFactory, value);
					String result = SerializeUtils.termToString(serializer, term);
					isOk = result.equals(value);
					FactoryUtils.returnTerm(term);
					if (isVerbose || !isOk) {
						out.print("PrefixedSerializer ");
						out.print(isOk ? "OK" : "KO");
						out.print(": ");
						out.print(value);
						out.print(" -> ");
						out.println(result);
					}
				}
				catch (ParsingException exception) {
					out.print("PrefixedSerializer ");
					out.println(value);
					out.print("PrefixedSerializer ");
					for (int k = 0; k < exception.getStringIndex(); k++)
						out.print(' ');
					out.println("^Error");
					isAllOk = false;
				}
				if (!isOk)
					isAllOk = false;
			}
			return isAllOk;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	////////////////////////////
	// Serializer implementation
	////////////////////////////

	public void writeTerm(PrintWriter out, Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf())
					out.print(this.leafChar);
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					out.print(this.endOfScopeChar);
					term = endOfScope.getBody();
					continue;
				}
				else if (term.isAbstraction()) {
					Abstraction abstraction = (Abstraction) term;
					out.print(this.abstractionChar);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					out.print(this.applicationChar);
					termStack.push(application.getArgument());
					term = application.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					return;
				term = termStack.pop();
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public void underlineTerm(
		PrintWriter out,
		Term rootTerm,
		Term underlinedTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			boolean isUnderlined = false;
			Term term = rootTerm;
			while (true) {
				if (term == underlinedTerm)
					isUnderlined = true;
				out.print(isUnderlined ? this.underlineChar : this.spaceChar);
				if (term.isLeaf()) {
					if (term == underlinedTerm)
						isUnderlined = false;
				}
				else {
					TermWithBody termWithBody = (TermWithBody) term;
					termStack.push(termWithBody);
					term = termWithBody.getBody();
					continue;
				}
				while (true) {
					if (termStack.isEmpty())
						return;
					Term child = term;
					term = termStack.pop();
					if (term.isApplication()) {
						Application application = (Application) term;
						if (child == application.getBody()) {
							termStack.push(application);
							term = application.getArgument();
							break;
						}
					}
					if (term == underlinedTerm)
						isUnderlined = false;
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public void underlineRedex(
		PrintWriter out,
		Term rootTerm,
		Application underlinedRedex) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			int underline = UNDERLINE_OUT;
			Abstraction redexAbstraction = null;
			{
				Term body = underlinedRedex.getBody();
				while (body.isEndOfScope())
					body = ((EndOfScope) body).getBody();
				if (body.isAbstraction())
					redexAbstraction = (Abstraction) body;
			}
			int abstractionIndex = -1;
			int redexAbstractionIndex = -1;
			Term term = rootTerm;
			while (true) {
				if (term == underlinedRedex)
					underline = redexAbstraction != null ? UNDERLINE_REDEX : UNDERLINE_IN;
				else if (underline == UNDERLINE_REDEX) {
					if (term.isEndOfScope())
						underline = UNDERLINE_REDEX_END_OF_SCOPE;
					else if (term.isAbstraction()) {
						underline = UNDERLINE_REDEX_ABSTRACTION;
						redexAbstractionIndex = abstractionIndex + 1;
					}
					else
						throw new IllegalStateException();
				}
				else if (underline == UNDERLINE_REDEX_END_OF_SCOPE) {
					if (term.isAbstraction()) {
						underline = UNDERLINE_REDEX_ABSTRACTION;
						redexAbstractionIndex = abstractionIndex + 1;
					}
				}
				else if (underline == UNDERLINE_REDEX_ABSTRACTION)
					underline = UNDERLINE_REDEX_FUNCTION;
				char ch = this.underlineChars[underline];
				if (term.isLeaf()) {
					out.print(
						abstractionIndex == redexAbstractionIndex
							? this.redexBindingChar
							: ch);
				}
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					out.print(
						abstractionIndex == redexAbstractionIndex
							? this.redexBindingChar
							: ch);
					abstractionIndex--;
					termStack.push(endOfScope);
					term = endOfScope.getBody();
					continue;
				}
				else if (term.isAbstraction()) {
					abstractionIndex++;
					Abstraction abstraction = (Abstraction) term;
					out.print(ch);
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					out.print(ch);
					termStack.push(application);
					term = application.getBody();
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty())
						return;
					Term child = term;
					term = termStack.pop();
					if (term.isEndOfScope())
							abstractionIndex++;
					else if (term.isAbstraction()) {
						if (redexAbstractionIndex == abstractionIndex)
							redexAbstractionIndex = -1;
						abstractionIndex--;
					}
					else if (term.isApplication()) {
						Application application = (Application) term;
						if (child == application.getBody()) {
							if (application == underlinedRedex)
								underline = UNDERLINE_REDEX_ARGUMENT;
							termStack.push(application);
							term = application.getArgument();
							break;
						}
						else {
							if (application == underlinedRedex)
								underline = UNDERLINE_OUT;
						}
					}
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public Term readTerm(Reader reader, TermFactory termFactory)
		throws IOException {
		if (reader == null || termFactory == null)
			throw new IllegalArgumentException();
		ParsingInput in = new ParsingInput(reader);
		TermStack termStack = termFactory.createTermStack(false);
		try {
			while (true) {
				Term term = null;
				char ch = in.getChar();
				in.next();
				if (ch == this.leafChar) {
					Leaf leaf = termFactory.createLeaf();
					term = leaf;
				}
				else if (ch == this.endOfScopeChar) {
					EndOfScope endOfScope = termFactory.createEndOfScope();
					termStack.push(endOfScope);
					continue;
				}
				else if (ch == this.abstractionChar) {
					Abstraction abstraction = termFactory.createAbstraction();
					termStack.push(abstraction);
					continue;
				}
				else if (ch == this.applicationChar) {
					Application application = termFactory.createApplication();
					termStack.push(application);
					continue;
				}
				else
					throw new ParsingException(in.getLocation() - 1);
				while (true) {
					if (termStack.isEmpty())
						return term;
					Term child = term;
					term = termStack.pop();
					if (term.isEndOfScope()) {
						EndOfScope endOfScope = (EndOfScope) term;
						endOfScope.setBody(child);
					}
					else if (term.isAbstraction()) {
						Abstraction abstraction = (Abstraction) term;
						abstraction.setBody(child);
					}
					else if (term.isApplication()) {
						Application application = (Application) term;
						if (application.getBody() == null) {
							application.setBody(child);
							termStack.push(application);
							break;
						}
						else
							application.setArgument(child);
					}
					else
						throw new IllegalStateException();
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}
}
