/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.serialize;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;

import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;

/**
 * Utilities about serialization.
 */
public abstract class SerializeUtils {

	/**
	 * Parse a String expression.
	 * @param serializer a Serializer.
	 * @param expressionFactory an ExpressionFactory.
	 * @param value a String representing an expression.
	 * @return the Expression.
	 */
	public static final Term stringToTerm(
		Serializer serializer,
		TermFactory termFactory,
		String value) {
		try {
			return serializer.readTerm(new StringReader(value), termFactory);
		}
		catch (IOException exception) {
			return null;
		}
	}

	/**
	 * Produces a String from a Term.
	 * @param serializer a Serializer.
	 * @param term a Term to serialize.
	 * @return a String representing the given Term.
	 */
	public static final String termToString(Serializer serializer, Term term) {
		StringWriter stringWriter = new StringWriter();
		serializer.writeTerm(new PrintWriter(stringWriter), term);
		return stringWriter.toString();
	}

	/**
	 * Print a Term on the standard output.
	 * @param serializer a Serializer.
	 * @param term a Term.
	 */
	public static final void printTerm(Serializer serializer, Term term) {
		printUnderlinedTerm(serializer, term, null);
	}

	/**
	 * Print an underlined Term on the standard output.
	 * @param serializer a Serializer.
	 * @param term a Term. May be null if the underlinedTerm is not null.
	 * @param underlinedTerm a sub-Term to underline. May be null.
	 */
	public static final void printUnderlinedTerm(
		Serializer serializer,
		Term term,
		Term underlinedTerm) {
		PrintWriter out = new PrintWriter(System.out);
		printUnderlinedTerm(out, serializer, term, underlinedTerm);
	}

	/**
	 * Print an underlined Term.
	 * @param a PrintWriter.
	 * @param serializer a Serializer.
	 * @param term a Term. May be null if the underlinedTerm is not null.
	 * @param underlinedTerm a sub-Term to underline. May be null.
	 */
	public static final void printUnderlinedTerm(
		PrintWriter out,
		Serializer serializer,
		Term term,
		Term underlinedTerm) {
		if (term == null) {
			if (underlinedTerm == null)
				throw new IllegalArgumentException();
			term = rootTerm(underlinedTerm);
		}
		serializer.writeTerm(out, term);
		out.println();
		if (underlinedTerm != null) {
			serializer.underlineTerm(out, term, underlinedTerm);
			out.println();
		}
		out.flush();
	}

	private static final Term rootTerm(final Term term) {
		Term root = term;
		while (true) {
			Term parent = root.getParent();
			if (parent == null)
				return root;
			root = parent;
		}
	}
}
