/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.serialize;

import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.lambda.tree.exception.ParsingException;
import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.model.TermWithBody;
import org.xmloperator.lambda.tree.util.FactoryUtils;

/**
 * Usual Expression serializer, convenient for a human reader.
 * <p>--- Term syntax</p>
 * <pre>
 * Term ::= Variable | EndOfScope | Abstraction | Application | '(' Term ')'
 * Variable ::= ( 'a' | ... | 'z' ) Number?
 * Number ::= ('0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9')+
 * Abstraction ::= '\' VariableList '.' Term
 * EndOfScope ::= '/' VariableList '.' Term
 * VariableList ::= Variable+
 * Application ::= Term Term
 * </pre>
 */
public class UsualSerializer implements Serializer {
	static final char SPACE_DEFAULT_CHAR = ' ';
	static final char UNDERLINE_DEFAULT_CHAR = '=';
	static final char REDEX_DEFAULT_CHAR = '-';
	static final char REDEX_ABSTRACTION_DEFAULT_CHAR = '+';
	static final char REDEX_BINDING_DEFAULT_CHAR = '^';
	static final char REDEX_ARGUMENT_DEFAULT_CHAR = '*';
	static final int UNDERLINE_OUT = 0;
	static final int UNDERLINE_IN = 1;
	static final int UNDERLINE_REDEX = 2;
	static final int UNDERLINE_REDEX_END_OF_SCOPE = 3;
	static final int UNDERLINE_REDEX_ABSTRACTION = 4;
	static final int UNDERLINE_REDEX_FUNCTION = 5;
	static final int UNDERLINE_REDEX_ARGUMENT = 6;

	private final char spaceChar = SPACE_DEFAULT_CHAR;
	private final char underlineChar = UNDERLINE_DEFAULT_CHAR;
	private final char redexChar = REDEX_DEFAULT_CHAR;
	private final char redexAbstractionChar = REDEX_ABSTRACTION_DEFAULT_CHAR;
	private final char redexBindingChar = REDEX_BINDING_DEFAULT_CHAR;
	private final char redexArgumentChar = REDEX_ARGUMENT_DEFAULT_CHAR;
	private final char[] underlineChars;

	private static final char[] SIMPLE_IDENTIFIERS =
		{ 'x', 'y', 'z', 'u', 'v', 'w', 'p', 'q', 'r', 's' };
	private static final char NUMBERED_IDENTIFIER = 't';
	private static final char ABSTRACTION_CHAR = '\\';
	private static final char END_OF_SCOPE_CHAR = '/';

	private final boolean isEndOfScopeWritenOut;

	public UsualSerializer(boolean isEndOfScopeWritenOut) {
		this.isEndOfScopeWritenOut = isEndOfScopeWritenOut;
		this.underlineChars =
			new char[] {
				this.spaceChar,
				this.redexChar,
				this.redexChar,
				this.redexChar,
				this.redexChar,
				this.redexChar,
				this.redexArgumentChar };
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final String VALUE_I = "\\x.x";
	private static final String VALUE_K = "\\xy./y.x";
	private static final String VALUE_F = "\\x./x.\\x.x";
	private static final String VALUE_C = "\\xyz.z(/zy.x)(/z.y)";
	private static final String VALUE_FST = "\\x.x(/x." + VALUE_K + ")";
	private static final String VALUE_SND = "\\x.x(/x." + VALUE_F + ")";
	private static final String VALUE_DELTA = "\\x.xx";
	private static final String VALUE_OMEGA =
		"(" + VALUE_DELTA + ")(" + VALUE_DELTA + ")";
	private static final String VALUE_Y = "\\x.(\\y.(/y.x)(yy))(\\y.(/y.x)(yy))";
	private static final String VALUE_V = "\\xy.y((/y.xx)y)";
	private static final String VALUE_TETA = "(" + VALUE_V + ")(" + VALUE_V + ")";

	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Serializes in and out some terms.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			String[] values =
				{
					VALUE_I,
					VALUE_K,
					VALUE_F,
					VALUE_C,
					VALUE_FST,
					VALUE_SND,
					VALUE_SND,
					VALUE_OMEGA,
					VALUE_Y,
					VALUE_Y,
					VALUE_TETA };
			TermFactory expressionFactory = TERM_FACTORY;
			boolean isAllOk = true;
			for (int index = 0; index < values.length; index++) {
				boolean isOk = false;
				String value = values[index];
				try {
					Serializer serializer = new UsualSerializer(true);
					Term term =
						SerializeUtils.stringToTerm(serializer, expressionFactory, value);
					String result = SerializeUtils.termToString(serializer, term);
					isOk = result.equals(value);
					FactoryUtils.returnTerm(term);
					if (isVerbose || !isOk) {
						out.print("UsualSerializer ");
						out.print(isOk ? "OK" : "KO");
						out.print(": ");
						out.print(value);
						out.print(" -> ");
						out.println(result);
					}
				}
				catch (ParsingException exception) {
					out.print("UsualSerializer ");
					out.println(value);
					out.print("UsualSerializer ");
					for (int k = 0; k < exception.getStringIndex(); k++)
						out.print(' ');
					out.println("^Error");
					isAllOk = false;
				}
				if (!isOk)
					isAllOk = false;
			}
			return isAllOk;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	private void writeIdentifier(PrintWriter out, int abstractionIndex) {
		if (abstractionIndex < 0)
			out.print('?');
		else if (abstractionIndex < SIMPLE_IDENTIFIERS.length)
			out.print(SIMPLE_IDENTIFIERS[abstractionIndex]);
		else {
			out.print(NUMBERED_IDENTIFIER);
			out.print(Integer.toString(abstractionIndex - SIMPLE_IDENTIFIERS.length));
		}
	}

	private void underlineIdentifier(
		PrintWriter out,
		int abstractionIndex,
		char firstChar,
		char nextChar) {
		out.print(firstChar);
		if (abstractionIndex >= SIMPLE_IDENTIFIERS.length)
			underlineNumber(
				out,
				abstractionIndex - SIMPLE_IDENTIFIERS.length,
				nextChar);
	}

	private static final void underlineNumber(
		PrintWriter out,
		int index,
		char ch) {
		out.print(ch);
		while (index > 9) {
			out.print(ch);
			index /= 10;
		}
	}

	private final boolean isFunctionParenthesed(Term function) {
		return function.isAbstraction()
			|| (function.isEndOfScope() && this.isEndOfScopeWritenOut);
	}

	private final boolean isArgumentParenthesed(Term argument) {
		return argument.isAbstraction()
			|| argument.isApplication()
			|| (argument.isEndOfScope() && this.isEndOfScopeWritenOut);
	}

	////////////////////////////
	// Serializer implementation
	////////////////////////////

	public void writeTerm(PrintWriter out, Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			int abstractionIndex = -1;
			while (true) {
				if (term.isLeaf())
					writeIdentifier(out, abstractionIndex);
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					if (this.isEndOfScopeWritenOut) {
						{
							Term parent = endOfScope.getParent();
							if (parent == null || !parent.isEndOfScope())
								out.print(END_OF_SCOPE_CHAR);
						}
						writeIdentifier(out, abstractionIndex);
						if (!endOfScope.getBody().isEndOfScope())
							out.print('.');
					}
					abstractionIndex--;
					termStack.push(endOfScope);
					term = endOfScope.getBody();
					continue;
				}
				else if (term.isAbstraction()) {
					abstractionIndex++;
					Abstraction abstraction = (Abstraction) term;
					{
						Term parent = abstraction.getParent();
						if (parent == null || !parent.isAbstraction())
							out.print(ABSTRACTION_CHAR);
					}
					writeIdentifier(out, abstractionIndex);
					if (!abstraction.getBody().isAbstraction())
						out.print('.');
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getBody();
					if (this.isFunctionParenthesed(term))
						out.print('(');
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty())
						return;
					Term child = term;
					term = termStack.pop();
					if (term.isEndOfScope())
						abstractionIndex++;
					else if (term.isAbstraction())
						abstractionIndex--;
					else if (term.isApplication()) {
						Application application = (Application) term;
						if (child == application.getBody()) {
							termStack.push(application);
							term = application.getArgument();
							if (this.isFunctionParenthesed(child))
								out.print(')');
							if (this.isArgumentParenthesed(term))
								out.print('(');
							break;
						}
						else {
							if (this.isArgumentParenthesed(child))
								out.print(')');
						}
					}
					else
						throw new IllegalStateException();
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public void underlineTerm(
		PrintWriter out,
		Term rootTerm,
		Term underlinedTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			int abstractionIndex = -1;
			char ch = this.spaceChar;
			while (true) {
				if (term == underlinedTerm)
					ch = this.underlineChar;
				if (term.isLeaf()) {
					underlineIdentifier(out, abstractionIndex, ch, ch);
					if (term == underlinedTerm)
						ch = this.spaceChar;
				}
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					if (this.isEndOfScopeWritenOut) {
						{
							Term parent = endOfScope.getParent();
							if (parent == null || !parent.isEndOfScope())
								out.print(ch);
						}
						underlineIdentifier(out, abstractionIndex, ch, ch);
						if (!endOfScope.getBody().isEndOfScope())
							out.print(ch);
					}
					abstractionIndex--;
					termStack.push(endOfScope);
					term = endOfScope.getBody();
					continue;
				}
				else if (term.isAbstraction()) {
					abstractionIndex++;
					Abstraction abstraction = (Abstraction) term;
					{
						Term parent = abstraction.getParent();
						if (parent == null || !parent.isAbstraction())
							out.print(ch);
					}
					underlineIdentifier(out, abstractionIndex, ch, ch);
					if (!abstraction.getBody().isAbstraction())
						out.print(ch);
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getBody();
					if (this.isFunctionParenthesed(term))
						out.print(ch);
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty())
						return;
					Term child = term;
					term = termStack.pop();
					if (term.isEndOfScope())
						abstractionIndex++;
					else if (term.isAbstraction())
						abstractionIndex--;
					else if (term.isApplication()) {
						Application application = (Application) term;
						if (child == application.getBody()) {
							termStack.push(application);
							term = application.getArgument();
							if (this.isFunctionParenthesed(child))
								out.print(ch);
							if (this.isArgumentParenthesed(term))
								out.print(ch);
							break;
						}
						else {
							if (this.isArgumentParenthesed(child))
								out.print(ch);
						}
					}
					else
						throw new IllegalStateException();
					if (term == underlinedTerm) {
						ch = this.spaceChar;
					}
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public void underlineRedex(
		PrintWriter out,
		Term rootTerm,
		Application underlinedRedex) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			int underline = UNDERLINE_OUT;
			Abstraction redexAbstraction = null;
			{
				Term body = underlinedRedex.getBody();
				while (body.isEndOfScope())
					body = ((EndOfScope) body).getBody();
				if (body.isAbstraction())
					redexAbstraction = (Abstraction) body;
			}
			Term term = rootTerm;
			int abstractionIndex = -1;
			int redexAbstractionIndex = -1;
			while (true) {
				if (term == underlinedRedex)
					underline = redexAbstraction != null ? UNDERLINE_REDEX : UNDERLINE_IN;
				else if (underline == UNDERLINE_REDEX) {
					if (term.isEndOfScope())
						underline = UNDERLINE_REDEX_END_OF_SCOPE;
					else if (term.isAbstraction()) {
						underline = UNDERLINE_REDEX_ABSTRACTION;
						redexAbstractionIndex = abstractionIndex + 1;
					}
					else
						throw new IllegalStateException();
				}
				else if (underline == UNDERLINE_REDEX_END_OF_SCOPE) {
					if (term.isAbstraction()) {
						underline = UNDERLINE_REDEX_ABSTRACTION;
						redexAbstractionIndex = abstractionIndex + 1;
					}
				}
				else if (underline == UNDERLINE_REDEX_ABSTRACTION)
					underline = UNDERLINE_REDEX_FUNCTION;
				char ch = this.underlineChars[underline];
				if (term.isLeaf()) {
					underlineIdentifier(
						out,
						abstractionIndex,
						abstractionIndex == redexAbstractionIndex
							? this.redexBindingChar
							: ch,
						ch);
				}
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					if (this.isEndOfScopeWritenOut) {
						{
							Term parent = endOfScope.getParent();
							if (parent == null || !parent.isEndOfScope())
								out.print(ch);
						}
						underlineIdentifier(
							out,
							abstractionIndex,
							abstractionIndex == redexAbstractionIndex
								? this.redexBindingChar
								: ch,
							ch);
						if (!endOfScope.getBody().isEndOfScope())
							out.print(ch);
					}
					abstractionIndex--;
					termStack.push(endOfScope);
					term = endOfScope.getBody();
					continue;
				}
				else if (term.isAbstraction()) {
					abstractionIndex++;
					Abstraction abstraction = (Abstraction) term;
					{
						Term parent = abstraction.getParent();
						if (parent == null || !parent.isAbstraction())
							out.print(ch);
					}
					underlineIdentifier(
						out,
						abstractionIndex,
						underline == UNDERLINE_REDEX_ABSTRACTION
							? this.redexAbstractionChar
							: ch,
						ch);
					if (!abstraction.getBody().isAbstraction())
						out.print(ch);
					termStack.push(abstraction);
					term = abstraction.getBody();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getBody();
					if (this.isFunctionParenthesed(term))
						out.print(ch);
					continue;
				}
				else
					throw new IllegalArgumentException();
				while (true) {
					if (termStack.isEmpty())
						return;
					Term child = term;
					term = termStack.pop();
					if (term.isEndOfScope())
						abstractionIndex++;
					else if (term.isAbstraction()) {
						if (redexAbstractionIndex == abstractionIndex)
							redexAbstractionIndex = -1;
						abstractionIndex--;
					}
					else if (term.isApplication()) {
						Application application = (Application) term;
						if (child == application.getBody()) {
							termStack.push(application);
							term = application.getArgument();
							if (this.isFunctionParenthesed(child))
								out.print(ch);
							if (application == underlinedRedex) {
								underline = UNDERLINE_REDEX_ARGUMENT;
								ch = this.redexArgumentChar;
							}
							if (this.isArgumentParenthesed(term))
								out.print(ch);
							break;
						}
						else {
							if (this.isArgumentParenthesed(child))
								out.print(ch);
							if (application == underlinedRedex) {
								underline = UNDERLINE_OUT;
								ch = this.spaceChar;
							}
						}
					}
				}
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}

	public Term readTerm(Reader reader, TermFactory termFactory)
		throws IOException {
		if (reader == null || termFactory == null)
			throw new IllegalArgumentException();
		final ParsingInput in = new ParsingInput(reader);
		final List identifierStack = new ArrayList();
		final List identifierUnStack = new ArrayList();
		final TermStack termStack = termFactory.createTermStack(false);
		final Term parenthesis = termFactory.createLeaf();
		try {
			Term term = null;
			while (true) {
				if (in.isEmpty()) {
					if (term == null)
						throw new ParsingException(in.getLocation());
					while (!termStack.isEmpty()) {
						Term child = term;
						term = termStack.pop();
						if (term == parenthesis)
							throw new ParsingException(in.getLocation());
						if (!term.isAbstraction() && !term.isEndOfScope())
							throw new IllegalStateException();
						((TermWithBody) term).setBody(child);
					}
					return term;
				}
				char ch = in.getChar();
				if (term != null && ch != ')') {
					Application application = termFactory.createApplication();
					application.setBody(term);
					termStack.push(application);
					term = null;
				}
				switch (ch) {
					case '(' :
						in.next();
						termStack.push(parenthesis);
						break;
					case ')' :
						if (term == null)
							throw new ParsingException(in.getLocation());
						while (true) {
							if (termStack.isEmpty())
								throw new ParsingException(in.getLocation());
							Term child = term;
							term = termStack.pop();
							if (term == parenthesis) {
								term = child;
								break;
							}
							if (term.isEndOfScope()) {
								((EndOfScope) term).setBody(child);
								identifierStack.add(
									identifierUnStack.remove(identifierUnStack.size() - 1));
							}
							else if (term.isAbstraction()) {
								((Abstraction) term).setBody(child);
								identifierStack.remove(identifierStack.size() - 1);
							}
							else
								throw new IllegalStateException();
						}
						in.next();
						break;
					case ABSTRACTION_CHAR :
						in.next();
						while (true) {
							identifierStack.add(in.parseIdentifier());
							termStack.push(termFactory.createAbstraction());
							if (!in.isIdentifier())
								break;
						}
						in.parseSeparator('.');
						break;
					case END_OF_SCOPE_CHAR :
						in.next();
						while (true) {
							int location = in.getLocation();
							String identifier = in.parseIdentifier();
							while (true) {
								if (identifierStack.isEmpty())
									throw new ParsingException(location);
								String expectedIdentifier =
									(String) identifierStack.remove(identifierStack.size() - 1);
								identifierUnStack.add(expectedIdentifier);
								termStack.push(termFactory.createEndOfScope());
								if (expectedIdentifier.equals(identifier))
									break;
							}
							if (!in.isIdentifier())
								break;
						}
						in.parseSeparator('.');
						break;
					default :
						{
							int location = in.getLocation();
							String identifier = in.parseIdentifier();
							int abstractionIndex = identifierStack.indexOf(identifier);
							if (abstractionIndex < 0)
								throw new ParsingException(location);
							int endOfScopeCount =
								identifierStack.size() - 1 - abstractionIndex;
							term = termFactory.createLeaf();
							for (int k = 0; k < endOfScopeCount; k++) {
								EndOfScope endOfScope = termFactory.createEndOfScope();
								endOfScope.setBody(term);
								term = endOfScope;
							}
						}
				}
				if (term != null
					&& !termStack.isEmpty()
					&& termStack.top().isApplication()) {
					Application application = (Application) termStack.pop();
					application.setArgument(term);
					term = application;
				}
			}
		}
		finally {
			termStack.returnToFactory();
			parenthesis.returnToFactory();
		}
	}
}
