/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.util;

import java.io.PrintStream;

import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;

/**
 * Common Terms. They are introduced in comments by their prefixed serialization
 *   form.
 */
public abstract class CommonTerms {

	/**
	 * Creates and returns an <tt>Abstraction</tt> that maps to <tt>"av"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newIdentity(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(termFactory.createLeaf());
	}

	/**
	 * Returns <tt>true</tt> if the given Abstraction is an Identity.
	 * @param abstraction an Abstraction.
	 * @return <tt>true</tt> if the given Abstraction is an Identity.
	 */
	public static final boolean isIdentity(Abstraction abstraction) {
		return abstraction.getBody().isLeaf();
	}

	/**
	 * Creates and returns an Abstraction that is immediatly followed by an
	 *   EndOfScope on itself.
	 * @param body the body of the EndOfScope.
	 * @return the created Abstraction.
	 * @throws NullPointerException if the given body is null.
	 */
	public static final Abstraction newNullAbstraction(Term body) {
		return FactoryUtils.newAbstraction(FactoryUtils.newEndOfScope(body));
	}

	/**
	 * Creates and returns an Abstraction that maps to K = <tt>"aaev"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newTrue(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			newNullAbstraction(termFactory.createLeaf()));
	}

	/**
	 * Creates and returns an Abstraction that maps to F = <tt>"aav"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newFalse(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(newIdentity(termFactory));
	}

	/**
	 * Creates and returns an Abstraction that, depending on value, maps to
	 *   value ? K : F.
	 * @param termFactory a TermFactory for building the Term.
	 * @param value a boolean value.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newBoolean(
		TermFactory termFactory,
		boolean value) {
		return value ? newTrue(termFactory) : newFalse(termFactory);
	}

	/**
	 * Creates and returns an Abstraction that maps to <tt>"aaaccveevev"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newCouple(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newAbstraction(
					FactoryUtils.newApplication(
						FactoryUtils.newApplication(
							termFactory.createLeaf(),
							FactoryUtils.newLeaf(termFactory, 2)),
						FactoryUtils.newLeaf(termFactory, 1)))));
	}

	/**
	 * Creates and returns an Abstraction that, depending on value, maps to
	 *   value ? "acvK" : "acvF".
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newFirstInCouple(
		TermFactory termFactory,
		boolean value) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newApplication(
				termFactory.createLeaf(),
				newBoolean(termFactory, value)));
	}

	/**
	 * Creates and returns an Application that maps to <tt>"cvv"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Application.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Application newLoop(TermFactory termFactory) {
		return FactoryUtils.newApplication(
			termFactory.createLeaf(),
			termFactory.createLeaf());
	}

	/**
	 * Return true if the given Application maps to <tt>"cvv"</tt>.
	 * @param application an Application.
	 * @return true if the given Application maps to <tt>"cvv"</tt>.
	 */
	public static final boolean isLoop(Application application) {
		return application.getBody().isLeaf()
			&& application.getArgument().isLeaf();
	}

	/**
	 * Creates and returns an Abstraction that maps to <tt>"D" = "acvv"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newDelta(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(newLoop(termFactory));
	}

	/**
	 * Return true if the given Abstraction is a D Expression.
	 * @param abstraction an Abstraction.
	 * @return true if it is a D Expression.
	 */
	public static final boolean isDelta(Abstraction abstraction) {
		Term body = abstraction.getBody();
		return body.isApplication() && isLoop((Application) body);
	}

	/**
	 * Creates and returns an Application that maps to <tt>"cDD"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Application.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Application newOmega(TermFactory termFactory) {
		return FactoryUtils.newApplication(
			newDelta(termFactory),
			newDelta(termFactory));
	}

	/**
	 * Return true if the given Application is an Omega Term.
	 * @param application an Application.
	 * @return true if it is an Omega Term.
	 */
	public static final boolean isOmega(Application application) {
		Term function = application.getBody();
		Term argument = application.getArgument();
		return function.isAbstraction()
			&& argument.isAbstraction()
			&& isDelta((Abstraction) function)
			&& isDelta((Abstraction) argument);
	}

	/**
	 * Creates and returns an Abstraction that maps to <tt>"acevcvv"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given termFactory is null.
	 */
	public static final Abstraction newYPart(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newApplication(
				FactoryUtils.newLeaf(termFactory, 1),
				newLoop(termFactory)));
	}

	/**
	 * Creates and returns an Abstraction that maps to <tt>"acacevcvvacevcvv"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newY(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newApplication(
				newYPart(termFactory),
				newYPart(termFactory)));
	}

	/**
	 * Creates and returns an Abstraction that maps to "V" = <tt>"aacvcecvvv"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Abstraction.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Abstraction newV(TermFactory termFactory) {
		return FactoryUtils.newAbstraction(
			FactoryUtils.newAbstraction(
				FactoryUtils.newApplication(
					termFactory.createLeaf(),
					FactoryUtils.newApplication(
						FactoryUtils.newEndOfScope(newLoop(termFactory)),
						termFactory.createLeaf()))));
	}

	/**
	 * Creates and returns an Application that maps to <tt>"cVV"</tt>.
	 * @param termFactory a TermFactory for building the Term.
	 * @return the created Application.
	 * @throws NullPOinterException if the given TermFactory is null.
	 */
	public static final Application newTeta(TermFactory termFactory) {
		return FactoryUtils.newApplication(
			newV(termFactory),
			newV(termFactory));
	}

	public static void main(String[] args) {
		if (test(System.out, true))
			System.out.println("All is OK");
		System.exit(0);
	}

	private static final TermFactory TERM_FACTORY = Test.TERM_FACTORY;
	private static final boolean IS_VERBOSE = false;

	/**
	 * Computes the size of some Expressions.
	 * @param out the PrintStream to print results. May be null.
	 * @param isVerbose if true then all tests have to be displayed.
	 * @return true if all is Ok.
	 */
	public static boolean test(PrintStream out, boolean isVerbose) {
		try {
			isVerbose |= IS_VERBOSE;
			TermFactory termFactory = TERM_FACTORY;
			boolean isOk = true;
			isOk &= test(out, isVerbose, newIdentity(termFactory), "I = av", 2);
			isOk
				&= test(
					out,
					isVerbose,
					newBoolean(termFactory, true),
					"K = aaev",
					3);
			isOk
				&= test(
					out,
					isVerbose,
					newBoolean(termFactory, false),
					"F = aav",
					3);
			isOk
				&= test(
					out,
					isVerbose,
					newCouple(termFactory),
					"C = aaaccveevev",
					6);
			isOk
				&= test(
					out,
					isVerbose,
					newFirstInCouple(termFactory, true),
					"Fst = acveK",
					5);
			isOk
				&= test(
					out,
					isVerbose,
					newFirstInCouple(termFactory, false),
					"Snd = acveF",
					5);
			isOk &= test(out, isVerbose, newDelta(termFactory), "D = acvv", 3);
			isOk &= test(out, isVerbose, newOmega(termFactory), "cDD", 6);
			isOk
				&= test(
					out,
					isVerbose,
					newY(termFactory),
					"Y = acacevcvvacevcvv",
					9);
			isOk
				&= test(out, isVerbose, newV(termFactory), "V = aacvcecvvv", 6);
			isOk &= test(out, isVerbose, newTeta(termFactory), "cVV", 12);
			return isOk;
		}
		catch (Exception exception) {
			exception.printStackTrace(out);
			return false;
		}
	}

	public static boolean test(
		PrintStream out,
		boolean isVerbose,
		Term term,
		String name,
		int size) {
		boolean isOk = SignatureUtils.leafAndAbstractionCount(term) == size;
		if (isVerbose || !isOk) {
			out.print("CommonsExpressions ");
			out.print(isOk ? "OK" : "KO");
			out.print(" Size(");
			out.print(name);
			out.print(") = ");
			out.println(SignatureUtils.leafAndAbstractionCount(term));
		}
		FactoryUtils.returnTerm(term);
		return isOk;
	}
}
