/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.util;

import org.xmloperator.lambda.tree.model.Abstraction;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.EndOfScope;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;

/**
 * Utilities about building Expressions and returning them to their factory.
 */
public abstract class FactoryUtils {

	/**
	 * Creates and returns a Leaf contained in a sequence of EndOfScopes.
	 * @param termFactory a TermFactory.
	 * @param endOfScopeCount a count of EndOfScope(s).
	 * @return the created Leaf.
	 */
	public static final Term newLeaf(
		TermFactory termFactory,
		int endOfScopeCount) {
		Term term = termFactory.createLeaf();
		for (int k = 0; k < endOfScopeCount; k++)
			term = newEndOfScope(term);
		return term;
	}

	/**
	 * Creates and returns a new EndOfScope.
	 * @param body a body for the EndOfSCope.
	 * @return the created EndOfScope.
	 * @throws NullPointerException if the given body is null.
	 */
	public static final EndOfScope newEndOfScope(Term body) {
		TermFactory termFactory = body.getFactory();
		EndOfScope endOfScope = termFactory.createEndOfScope();
		endOfScope.setBody(body);
		return endOfScope;
	}

	/**
	 * Creates and returns a new Abstraction.
	 * @param body a body for the Abstraction.
	 * @return the created Abstraction.
	 * @throws NullPointerException if the given body is null.
	 */
	public static final Abstraction newAbstraction(Term body) {
		TermFactory termFactory = body.getFactory();
		Abstraction abstraction = termFactory.createAbstraction();
		abstraction.setBody(body);
		return abstraction;
	}

	/**
	 * Creates and returns a new Application.
	 * @param body its body.
	 * @param argument its argument.
	 * @return the created Application.
	 * @throws NullPointerException if the given body is null.
	 */
	public static final Application newApplication(Term body, Term argument) {
		TermFactory termFactory = body.getFactory();
		Application application = termFactory.createApplication();
		application.setBody(body);
		application.setArgument(argument);
		return application;
	}

	/**
	 * Return a Term and its content to the factory.
	 * @param rootTerm a no more used Term.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final void returnTerm(Term rootTerm) {
		if (rootTerm.getParent() != null)
			throw new IllegalArgumentException();
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			while (true) {
				if (term.isLeaf())
					term.returnToFactory();
				else if (term.isEndOfScope()) {
					EndOfScope endOfScope = (EndOfScope) term;
					term = endOfScope.getBody();
					endOfScope.clearBody();
					endOfScope.returnToFactory();
					continue;
				}
				else if (term.isAbstraction()) {
					Abstraction abstraction = (Abstraction) term;
					term = abstraction.getBody();
					abstraction.clearBody();
					abstraction.returnToFactory();
					continue;
				}
				else if (term.isApplication()) {
					Application application = (Application) term;
					termStack.push(application);
					term = application.getArgument();
					application.clearArgument();
					continue;
				}
				else
					throw new IllegalArgumentException();
				if (termStack.isEmpty())
					break;
				Application application = (Application) termStack.pop();
				term = application.getBody();
				application.clearBody();
				application.returnToFactory();
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}
}
