/**
 * Copyright (c) 2004 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda/).
 */
package org.xmloperator.lambda.tree.util;

import org.xmloperator.lambda.tree.exception.BadFormedException;
import org.xmloperator.lambda.tree.model.Application;
import org.xmloperator.lambda.tree.model.Term;
import org.xmloperator.lambda.tree.model.TermFactory;
import org.xmloperator.lambda.tree.model.TermStack;
import org.xmloperator.lambda.tree.model.TermWithBody;

/**
 * Utilities about well-formedness of Terms.
 * In order to be well formed, a Term must conform to the following constraints :
 * <ul>
 * <li>The term cannot be broken, i.e. no body nor argument can be null.</li>
 * <li>Each sub-Term must be correctly linked to his parent.</li>
 * <li>There must exist no loop.</li>
 * </ul>
 */
public abstract class WellFormedness {

	/**
	 * Verify the well-formedness of a Term.
	 * @param rootTerm a Term to verify.
	 * @throws BadFormedException if the given Term doesn't
	 *   conform to the constraint 2 or 3.
	 * @throws NullPointerException if the given Term is broken.
	 */
	public static final void verifyWellFormedness(final Term rootTerm) {
		TermFactory termFactory = rootTerm.getFactory();
		TermStack termStack = termFactory.createTermStack(false);
		try {
			Term term = rootTerm;
			while (true) {
				if (termStack.contains(term))
					throw new BadFormedException(term, "Circular Term");
				if (!term.isLeaf()) {
					TermWithBody termWithBody = (TermWithBody) term;
					termStack.push(term);
					term = termWithBody.getBody();
					if (term.getParent() != termWithBody)
						throw new BadFormedException(term, "Contradictory body Term");
					continue;
				}
				while (true) {
					if (termStack.isEmpty()) {
						term = null;
						break;
					}
					Term child = term;
					term = termStack.pop();
					if (term.isApplication()) {
						Application application = (Application) term;
						if (child == application.getBody()) {
							termStack.push(application);
							term = application.getArgument();
							if (term.getParent() != application)
								throw new BadFormedException(
									term,
									"Contradictory argument Term");
							break;
						}
					}
				}
				if (term == null)
					break;
			}
		}
		finally {
			termStack.returnToFactory();
		}
	}
}
